"""Exceptions.

"""


class CompileError(Exception):
    """General ASN.1 compile error.

    """


class EncodeError(Exception):
    """General ASN.1 encode error.

    """


class DecodeError(Exception):
    """General ASN.1 decode error.

    """



EXPECTED = {'ModuleTagsImplicit': {'extensibility-implied': False,
                        'imports': {},
                        'object-classes': {},
                        'object-sets': {},
                        'tags': 'IMPLICIT',
                        'types': {'A': {'tag': {'kind': 'IMPLICIT',
                                                'number': 3},
                                        'type': 'INTEGER'},
                                  'AI': {'tag': {'kind': 'IMPLICIT',
                                                 'number': 3},
                                         'type': 'INTEGER'},
                                  'BA': {'tag': {'kind': 'IMPLICIT',
                                                 'number': 4},
                                         'type': 'A'},
                                  'BIA': {'tag': {'kind': 'IMPLICIT',
                                                  'number': 4},
                                          'type': 'A'},
                                  'BIAI': {'tag': {'kind': 'IMPLICIT',
                                                   'number': 4},
                                           'type': 'AI'},
                                  'C1': {'members': [{'members': [{'name': 'a',
                                                                   'tag': {'kind': 'IMPLICIT',
                                                                           'number': 0},
                                                                   'type': 'INTEGER'}],
                                                      'name': 'a',
                                                      'tag': {'kind': 'EXPLICIT',
                                                              'number': 0},
                                                      'type': 'CHOICE'}],
                                         'type': 'CHOICE'},
                                  'CBA': {'tag': {'kind': 'IMPLICIT',
                                                  'number': 5},
                                          'type': 'BA'},
                                  'CBIAI': {'tag': {'kind': 'IMPLICIT',
                                                    'number': 5},
                                            'type': 'BIAI'},
                                  'CIBIA': {'tag': {'kind': 'IMPLICIT',
                                                    'number': 5},
                                            'type': 'BIA'},
                                  'CIBIAI': {'tag': {'kind': 'IMPLICIT',
                                                     'number': 5},
                                             'type': 'BIAI'},
                                  'S1': {'members': [{'name': 'a',
                                                      'type': 'INTEGER'},
                                                     {'name': 'b',
                                                      'optional': True,
                                                      'type': 'BOOLEAN'}],
                                         'type': 'SEQUENCE'},
                                  'S2': {'members': [{'name': 'a',
                                                      'type': 'INTEGER'},
                                                     {'name': 'b',
                                                      'tag': {'kind': 'IMPLICIT',
                                                              'number': 2},
                                                      'type': 'S1'},
                                                     {'members': [{'name': 'a',
                                                                   'type': 'BOOLEAN'}],
                                                      'name': 'c',
                                                      'type': 'CHOICE'}],
                                         'type': 'SEQUENCE'},
                                  'S3': {'members': [{'name': 'a',
                                                      'type': 'INTEGER'},
                                                     {'name': 'b',
                                                      'tag': {'kind': 'IMPLICIT',
                                                              'number': 2},
                                                      'type': 'S1'},
                                                     {'members': [{'name': 'a',
                                                                   'type': 'BOOLEAN'}],
                                                      'name': 'c',
                                                      'tag': {'kind': 'EXPLICIT',
                                                              'number': 3},
                                                      'type': 'CHOICE'}],
                                         'type': 'SEQUENCE'},
                                  'S4': {'members': [{'name': 'a',
                                                      'type': 'INTEGER'},
                                                     {'name': 'b',
                                                      'tag': {'kind': 'EXPLICIT',
                                                              'number': 1},
                                                      'type': 'C1'},
                                                     {'name': 'c',
                                                      'tag': {'kind': 'IMPLICIT',
                                                              'number': 2},
                                                      'type': 'S1'},
                                                     {'members': [{'name': 'a',
                                                                   'type': 'BOOLEAN'}],
                                                      'name': 'd',
                                                      'type': 'CHOICE'}],
                                         'type': 'SEQUENCE'},
                                  'S5': {'members': [{'name': 'a',
                                                      'type': 'INTEGER'},
                                                     {'name': 'b',
                                                      'type': 'S1'},
                                                     {'members': [{'name': 'a',
                                                                   'type': 'BOOLEAN'}],
                                                      'name': 'c',
                                                      'type': 'CHOICE'}],
                                         'type': 'SEQUENCE'},
                                  'S6': {'members': [{'name': 'a',
                                                      'type': 'INTEGER'},
                                                     '...',
                                                     {'name': 'b',
                                                      'type': 'BOOLEAN'}],
                                         'type': 'SEQUENCE'},
                                  'S7': {'members': [{'name': 'a',
                                                      'tag': {'kind': 'IMPLICIT',
                                                              'number': 2},
                                                      'type': 'INTEGER'},
                                                     '...',
                                                     {'name': 'b',
                                                      'type': 'BOOLEAN'}],
                                          'type': 'SEQUENCE'},
                                   'S8': {'members': [{'element': {'members': [{'name': 'a',
                                                                                'type': 'INTEGER'},
                                                                               {'name': 'b',
                                                                                'type': 'BOOLEAN'}],
                                                                   'type': 'CHOICE'},
                                                       'name': 'a',
                                                       'type': 'SEQUENCE OF'}],
                                         'type': 'SEQUENCE'},
                                  'S9': {'members': [{'element': {'members': [{'name': 'a',
                                                                               'type': 'INTEGER'},
                                                                              {'name': 'b',
                                                                               'type': 'BOOLEAN'}],
                                                                  'type': 'CHOICE'},
                                                      'name': 'a',
                                                      'type': 'SET OF'}],
                                         'type': 'SET'}},
                        'values': {}}}



import unittest
from binascii import hexlify

class Asn1ToolsBaseTest(unittest.TestCase):

    def assert_encode_decode(self,
                             specification,
                             type_name,
                             decoded_message,
                             encoded_message):
        try:
            encoded = specification.encode(type_name, decoded_message)
        except:
            print('Failed to encode type with name {}.'.format(type_name))
            raise

        try:
            self.assertEqual(encoded, encoded_message)
        except:
            print('Wrong encoding of type with name {}.'.format(type_name))
            print('Actual:', hexlify(encoded))
            print('Expected:', hexlify(encoded_message))
            raise

        try:
            decoded = specification.decode(type_name, encoded_message)
        except:
            print('Failed to decode type with name {}.'.format(type_name))
            raise

        try:
            self.assertEqual(decoded, decoded_message)
        except:
            print('Wrong decoding of type with name {}.'.format(type_name))
            print('Actual:', decoded)
            print('Expected:', decoded_message)
            raise



EXPECTED = {'ExtensibilityImplied': {'extensibility-implied': True,
                          'imports': {},
                          'object-classes': {},
                          'object-sets': {},
                          'types': {'Choice': {'members': [{'name': 'a',
                                                            'type': 'INTEGER'},
                                                           '...'],
                                               'type': 'CHOICE'},
                                    'Choice2': {'members': [{'name': 'a',
                                                             'type': 'INTEGER'},
                                                            '...'],
                                                'type': 'CHOICE'},
                                    'Enum': {'values': {1: 'one',
                                                   2: 'two',
                                                   3: '...'},
                                             'type': 'ENUMERATED'},
                                    'Nested': {'members': [{'members': [{'members': [{'name': 'a',
                                                                                       'type': 'INTEGER'},
                                                                                      '...'],
                                                                          'name': 'b',
                                                                          'type': 'CHOICE'},
                                                                         '...'],
                                                             'name': 'a',
                                                             'type': 'SEQUENCE'},
                                                            '...'],
                                                'type': 'SET'},
                                    'Sequence': {'members': ['...'],
                                                 'type': 'SEQUENCE'},
                                    'Sequence2': {'members': [{'default': 0,
                                                               'name': 'a',
                                                               'type': 'INTEGER'},
                                                              '...'],
                                                  'type': 'SEQUENCE'},
                                    'Sequence3': {'members': [{'name': 'a',
                                                               'type': 'BOOLEAN'},
                                                              '...'],
                                                  'type': 'SEQUENCE'},
                                    'Sequence4': {'members': [{'name': 'a',
                                                               'type': 'BOOLEAN'},
                                                              '...',
                                                              {'name': 'b',
                                                               'type': 'BOOLEAN'}],
                                                  'type': 'SEQUENCE'},
                                    'Sequence5': {'members': [{'name': 'a',
                                                               'type': 'BOOLEAN'},
                                                              '...',
                                                              {'name': 'b',
                                                               'type': 'BOOLEAN'},
                                                              '...'],
                                                  'type': 'SEQUENCE'},
                                    'Sequence6': {'members': [{'name': 'a',
                                                               'type': 'BOOLEAN'},
                                                              '...',
                                                              {'name': 'b',
                                                               'type': 'BOOLEAN'},
                                                              '...',
                                                              {'name': 'c',
                                                               'type': 'BOOLEAN'}],
                                                  'type': 'SEQUENCE'},
                                    'Sequence7': {'members': [{'name': 'a',
                                                               'type': 'BOOLEAN'},
                                                              '...',
                                                              {'name': 'b',
                                                               'type': 'BOOLEAN'},
                                                              {'name': 'c',
                                                               'type': 'BOOLEAN'},
                                                              '...',
                                                              {'name': 'd',
                                                               'type': 'BOOLEAN'}],
                                                  'type': 'SEQUENCE'},
                                    'Sequence8': {'members': [{'name': 'a',
                                                               'type': 'BOOLEAN'},
                                                              '...',
                                                              '...'],
                                                  'type': 'SEQUENCE'},
                                    'Set': {'members': ['...'], 'type': 'SET'},
                                    'Set2': {'members': [{'default': 1,
                                                          'name': 'a',
                                                          'type': 'INTEGER'},
                                                         '...'],
                                             'type': 'SET'}},
                          'values': {}}}



import unittest
from .utils import Asn1ToolsBaseTest
import asn1tools


class Asn1ToolsTwoModulesTest(Asn1ToolsBaseTest):

    maxDiff = None

    def test_uper(self):
        files = [
            'tests/files/module-1.asn',
            'tests/files/module-2.asn'
        ]

        spec = asn1tools.compile_files(files, 'uper')

        # Message 1.
        decoded = {
            'priority': 'high',
            'src': 1,
            'dst': 2,
            'num': 0,
            'length': 256
        }

        encoded = b'\x14\x04\x00'

        self.assert_encode_decode(spec, 'ASequence', decoded, encoded)

        # Message 2.
        decoded = {
            'isvalide': True,
            'name': 'toto',
            'identity': 5
        }

        encoded = b'\x82\x74\xdf\xd3\x7d'

        self.assert_encode_decode(spec, 'SeqMod2', decoded, encoded)


if __name__ == '__main__':
    unittest.main()



EXPECTED = {'X691-A4': {'extensibility-implied': False,
             'imports': {},
             'object-classes': {},
             'object-sets': {},
             'tags': 'AUTOMATIC',
             'types': {'Ax': {'members': [{'name': 'a',
                                           'restricted-to': [(250, 253)],
                                           'type': 'INTEGER'},
                                          {'name': 'b', 'type': 'BOOLEAN'},
                                          {'members': [{'name': 'd',
                                                        'type': 'INTEGER'},
                                                       '...',
                                                       {'name': 'e',
                                                        'type': 'BOOLEAN'},
                                                       {'name': 'f',
                                                        'type': 'IA5String'},
                                                       '...'],
                                           'name': 'c',
                                           'type': 'CHOICE'},
                                          '...',
                                          {'name': 'g',
                                           'type': 'NumericString'},
                                          {'name': 'h',
                                           'optional': True,
                                           'type': 'BOOLEAN'},
                                          '...',
                                          {'name': 'i',
                                           'optional': True,
                                           'type': 'BMPString'},
                                          {'name': 'j',
                                           'optional': True,
                                           'type': 'PrintableString'}],
                              'type': 'SEQUENCE'}},
             'values': {}}}


import unittest
from .utils import Asn1ToolsBaseTest
import asn1tools
import sys
from copy import deepcopy

sys.path.append('tests/files')
sys.path.append('tests/files/3gpp')

from rrc_8_6_0 import EXPECTED as RRC_8_6_0
from lpp_14_3_0 import EXPECTED as LPP_14_3_0
from x691_a2 import EXPECTED as X691_A2
from x691_a4 import EXPECTED as X691_A4


class Asn1ToolsUPerTest(Asn1ToolsBaseTest):

    maxDiff = None

    def test_foo(self):
        foo = asn1tools.compile_files('tests/files/foo.asn', 'uper')

        self.assertEqual(len(foo.types), 2)
        self.assertTrue(foo.types['Question'] is not None)
        self.assertTrue(foo.types['Answer'] is not None)
        self.assertEqual(len(foo.modules), 1)
        self.assertTrue(foo.modules['Foo'] is not None)

        # Question.
        encoded = foo.encode('Question', {'id': 1, 'question': 'Is 1+1=3?'})
        self.assertEqual(encoded, b'\x01\x01\x09\x93\xcd\x03\x15\x6c\x5e\xb3\x7e')
        decoded = foo.decode('Question', encoded)
        self.assertEqual(decoded, {'id': 1, 'question': 'Is 1+1=3?'})

        # Answer.
        encoded = foo.encode('Answer', {'id': 1, 'answer': False})
        self.assertEqual(encoded, b'\x01\x01\x00')
        decoded = foo.decode('Answer', encoded)
        self.assertEqual(decoded, {'id': 1, 'answer': False})

        # Encode a question with missing field 'id'.
        with self.assertRaises(asn1tools.EncodeError) as cm:
            encoded = foo.encode('Question', {'question': 'Is 1+1=3?'})

        self.assertEqual(
            str(cm.exception),
            "Sequence member 'id' not found in {'question': 'Is 1+1=3?'}.")

    def test_decode_length(self):
        foo = asn1tools.compile_files('tests/files/foo.asn', 'uper')

        with self.assertRaises(asn1tools.DecodeError) as cm:
            foo.decode_length(b'')

        self.assertEqual(str(cm.exception),
                         ': Decode length not supported for this codec.')

    def test_x691_a1(self):
        a1 = asn1tools.compile_files('tests/files/x691_a1.asn', 'uper')

        decoded = {
            'name': {
                'givenName': 'John',
                'initial': 'P',
                'familyName': 'Smith'
            },
            'title': 'Director',
            'number': 51,
            'dateOfHire': '19710917',
            'nameOfSpouse': {
                'givenName': 'Mary',
                'initial': 'T',
                'familyName': 'Smith'
            },
            'children': [
                {
                    'name': {
                        'givenName': 'Ralph',
                        'initial': 'T',
                        'familyName': 'Smith'
                    },
                    'dateOfBirth': '19571111'
                },
                {
                    'name': {
                        'givenName': 'Susan',
                        'initial': 'B',
                        'familyName': 'Jones'
                    },
                    'dateOfBirth': '19590717'
                }
            ]
        }

        encoded = (
            b'\x82\x4a\xdf\xa3\x70\x0d\x00\x5a\x7b\x74\xf4\xd0\x02\x66\x11\x13'
            b'\x4f\x2c\xb8\xfa\x6f\xe4\x10\xc5\xcb\x76\x2c\x1c\xb1\x6e\x09\x37'
            b'\x0f\x2f\x20\x35\x01\x69\xed\xd3\xd3\x40\x10\x2d\x2c\x3b\x38\x68'
            b'\x01\xa8\x0b\x4f\x6e\x9e\x9a\x02\x18\xb9\x6a\xdd\x8b\x16\x2c\x41'
            b'\x69\xf5\xe7\x87\x70\x0c\x20\x59\x5b\xf7\x65\xe6\x10\xc5\xcb\x57'
            b'\x2c\x1b\xb1\x6e'
        )

        self.assert_encode_decode(a1, 'PersonnelRecord', decoded, encoded)

    def test_x691_a2(self):
        a2 = asn1tools.compile_dict(deepcopy(X691_A2), 'uper')

        decoded = {
            'name': {
                'givenName': 'John',
                'initial': 'P',
                'familyName': 'Smith'
            },
            'title': 'Director',
            'number': 51,
            'dateOfHire': '19710917',
            'nameOfSpouse': {
                'givenName': 'Mary',
                'initial': 'T',
                'familyName': 'Smith'
            },
            'children': [
                {
                    'name': {
                        'givenName': 'Ralph',
                        'initial': 'T',
                        'familyName': 'Smith'
                    },
                    'dateOfBirth': '19571111'
                },
                {
                    'name': {
                        'givenName': 'Susan',
                        'initial': 'B',
                        'familyName': 'Jones'
                    },
                    'dateOfBirth': '19590717'
                }
            ]
        }

        encoded = (
            b'\x86\x5d\x51\xd2\x88\x8a\x51\x25\xf1\x80\x99\x84\x44\xd3\xcb\x2e'
            b'\x3e\x9b\xf9\x0c\xb8\x84\x8b\x86\x73\x96\xe8\xa8\x8a\x51\x25\xf1'
            b'\x81\x08\x9b\x93\xd7\x1a\xa2\x29\x44\x97\xc6\x32\xae\x22\x22\x22'
            b'\x98\x5c\xe5\x21\x88\x5d\x54\xc1\x70\xca\xc8\x38\xb8'
        )

        self.assert_encode_decode(a2, 'PersonnelRecord', decoded, encoded)

    def test_x691_a3(self):
        with self.assertRaises(asn1tools.ParseError) as cm:
            a3 = asn1tools.compile_files('tests/files/x691_a3.asn', 'uper')

        self.assertEqual(
            str(cm.exception),
            "Invalid ASN.1 syntax at line 10, column 22: 'SEQUENCE >!<"
            "(SIZE(2, ...)) OF ChildInformation OPTIONAL,': Expected \"{\".")

        return

        decoded = {
            'name': {
                'givenName': 'John',
                'initial': 'P',
                'familyName': 'Smith'
            },
            'title': 'Director',
            'number': 51,
            'dateOfHire': '19710917',
            'nameOfSpouse': {
                'givenName': 'Mary',
                'initial': 'T',
                'familyName': 'Smith'
            },
            'children': [
                {
                    'name': {
                        'givenName': 'Ralph',
                        'initial': 'T',
                        'familyName': 'Smith'
                    },
                    'dateOfBirth': '19571111'
                },
                {
                    'name': {
                        'givenName': 'Susan',
                        'initial': 'B',
                        'familyName': 'Jones'
                    },
                    'dateOfBirth': '19590717',
                    'sex': 'female'
                }
            ]
        }

        encoded = (
            b'\x40\xcb\xaa\x3a\x51\x08\xa5\x12\x5f\x18\x03\x30\x88\x9a\x79\x65'
            b'\xc7\xd3\x7f\x20\xcb\x88\x48\xb8\x19\xce\x5b\xa2\xa1\x14\xa2\x4b'
            b'\xe3\x01\x13\x72\x7a\xe3\x54\x22\x94\x49\x7c\x61\x95\x71\x11\x18'
            b'\x22\x98\x5c\xe5\x21\x84\x2e\xaa\x60\xb8\x32\xb2\x0e\x2e\x02\x02'
            b'\x80'
        )

        self.assert_encode_decode(a3, 'PersonnelRecord', decoded, encoded)

    def test_x691_a4(self):
        a4 = asn1tools.compile_dict(deepcopy(X691_A4), 'uper')

        decoded = {
            'a': 253,
            'b': True,
            'c': {
                'e': True
            },
            'g': '123',
            'h': True
        }

        encoded = (
            b'\x9e\x00\x06\x00\x04\x0a\x46\x90'
        )

        with self.assertRaises(asn1tools.EncodeError):
            self.assert_encode_decode(a4, 'Ax', decoded, encoded)

    def test_rrc_8_6_0(self):
        rrc = asn1tools.compile_dict(deepcopy(RRC_8_6_0), 'uper')

        # Message 1.
        decoded = {
            'message': {
                'c1' : {
                    'paging': {
                        'systemInfoModification': 'true',
                        'nonCriticalExtension': {
                        }
                    }
                }
            }
        }

        encoded = b'\x28'

        self.assert_encode_decode(rrc, 'PCCH-Message', decoded, encoded)

        # Message 2.
        decoded = {
            'message': {
                'c1' : {
                    'paging': {
                    }
                }
            }
        }

        encoded = b'\x00'

        self.assert_encode_decode(rrc, 'PCCH-Message', decoded, encoded)

        # Message 3.
        decoded = {
            'message': {
                'dl-Bandwidth': 'n6',
                'phich-Config': {
                    'phich-Duration': 'normal',
                    'phich-Resource': 'half'
                },
                'systemFrameNumber': (b'\x12', 8),
                'spare': (b'\x34\x40', 10)
            }
        }

        encoded = b'\x04\x48\xd1'

        self.assert_encode_decode(rrc, 'BCCH-BCH-Message', decoded, encoded)

        # Message 4.
        decoded = {
            'message': {
                'c1': {
                    'systemInformation': {
                        'criticalExtensions': {
                            'systemInformation-r8': {
                                'sib-TypeAndInfo': [
                                    {
                                        'sib2': {
                                            'ac-BarringInfo': {
                                                'ac-BarringForEmergency': True,
                                                'ac-BarringForMO-Data': {
                                                    'ac-BarringFactor': 'p95',
                                                    'ac-BarringTime': 's128',
                                                    'ac-BarringForSpecialAC': (b'\xf0', 5)
                                                }
                                            },
                                            'radioResourceConfigCommon': {
                                                'rach-ConfigCommon': {
                                                    'preambleInfo': {
                                                        'numberOfRA-Preambles': 'n24',
                                                        'preamblesGroupAConfig': {
                                                            'sizeOfRA-PreamblesGroupA': 'n28',
                                                            'messageSizeGroupA': 'b144',
                                                            'messagePowerOffsetGroupB': 'minusinfinity'
                                                        }
                                                    },
                                                    'powerRampingParameters': {
                                                        'powerRampingStep': 'dB0',
                                                        'preambleInitialReceivedTargetPower': 'dBm-102'
                                                    },
                                                    'ra-SupervisionInfo': {
                                                        'preambleTransMax': 'n8',
                                                        'ra-ResponseWindowSize': 'sf6',
                                                        'mac-ContentionResolutionTimer': 'sf48'
                                                    },
                                                    'maxHARQ-Msg3Tx': 8
                                                },
                                                'bcch-Config': {
                                                    'modificationPeriodCoeff': 'n2'
                                                },
                                                'pcch-Config': {
                                                    'defaultPagingCycle': 'rf256',
                                                    'nB': 'twoT'
                                                },
                                                'prach-Config': {
                                                    'rootSequenceIndex': 836,
                                                    'prach-ConfigInfo': {
                                                        'prach-ConfigIndex': 33,
                                                        'highSpeedFlag': False,
                                                        'zeroCorrelationZoneConfig': 10,
                                                        'prach-FreqOffset': 64
                                                    }
                                                },
                                                'pdsch-ConfigCommon': {
                                                    'referenceSignalPower': -60,
                                                    'p-b': 2
                                                },
                                                'pusch-ConfigCommon': {
                                                    'pusch-ConfigBasic': {
                                                        'n-SB': 1,
                                                        'hoppingMode': 'interSubFrame',
                                                        'pusch-HoppingOffset': 10,
                                                        'enable64QAM': False
                                                    },
                                                    'ul-ReferenceSignalsPUSCH': {
                                                        'groupHoppingEnabled': True,
                                                        'groupAssignmentPUSCH': 22,
                                                        'sequenceHoppingEnabled': False,
                                                        'cyclicShift': 5
                                                    }
                                                },
                                                'pucch-ConfigCommon': {
                                                    'deltaPUCCH-Shift': 'ds1',
                                                    'nRB-CQI': 98,
                                                    'nCS-AN': 4,
                                                    'n1PUCCH-AN': 2047
                                                },
                                                'soundingRS-UL-ConfigCommon': {
                                                    'setup': {
                                                        'srs-BandwidthConfig': 'bw0',
                                                        'srs-SubframeConfig': 'sc4',
                                                        'ackNackSRS-SimultaneousTransmission': True
                                                    }
                                                },
                                                'uplinkPowerControlCommon': {
                                                    'p0-NominalPUSCH': -126,
                                                    'alpha': 'al0',
                                                    'p0-NominalPUCCH': -127,
                                                    'deltaFList-PUCCH': {
                                                        'deltaF-PUCCH-Format1': 'deltaF-2',
                                                        'deltaF-PUCCH-Format1b': 'deltaF1',
                                                        'deltaF-PUCCH-Format2': 'deltaF0',
                                                        'deltaF-PUCCH-Format2a': 'deltaF-2',
                                                        'deltaF-PUCCH-Format2b': 'deltaF0'
                                                    },
                                                    'deltaPreambleMsg3': -1
                                                },
                                                'ul-CyclicPrefixLength': 'len1'
                                            },
                                            'ue-TimersAndConstants': {
                                                't300': 'ms100',
                                                't301': 'ms200',
                                                't310': 'ms50',
                                                'n310': 'n2',
                                                't311': 'ms30000',
                                                'n311': 'n2'
                                            },
                                            'freqInfo': {
                                                'additionalSpectrumEmission': 3
                                            },
                                            'timeAlignmentTimerCommon': 'sf500'
                                        }
                                    },
                                    {
                                        'sib3': {
                                            'cellReselectionInfoCommon': {
                                                'q-Hyst': 'dB0',
                                                'speedStateReselectionPars': {
                                                    'mobilityStateParameters': {
                                                        't-Evaluation': 's180',
                                                        't-HystNormal': 's180',
                                                        'n-CellChangeMedium': 1,
                                                        'n-CellChangeHigh': 16
                                                    },
                                                    'q-HystSF': {
                                                        'sf-Medium': 'dB-6',
                                                        'sf-High': 'dB-4'
                                                    }
                                                }
                                            },
                                            'cellReselectionServingFreqInfo': {
                                                'threshServingLow': 7,
                                                'cellReselectionPriority': 3
                                            },
                                            'intraFreqCellReselectionInfo': {
                                                'q-RxLevMin': -33,
                                                's-IntraSearch': 0,
                                                'presenceAntennaPort1': False,
                                                'neighCellConfig': (b'\x80', 2),
                                                't-ReselectionEUTRA': 4
                                            }
                                        }
                                    },
                                    {
                                        'sib4': {
                                        }
                                    },
                                    {
                                        'sib5': {
                                            'interFreqCarrierFreqList': [
                                                {
                                                    'dl-CarrierFreq': 1,
                                                    'q-RxLevMin': -45,
                                                    't-ReselectionEUTRA': 0,
                                                    'threshX-High': 31,
                                                    'threshX-Low': 29 ,
                                                    'allowedMeasBandwidth': 'mbw6',
                                                    'presenceAntennaPort1': True,
                                                    'q-OffsetFreq': 'dB0',
                                                    'neighCellConfig': (b'\x00', 2)
                                                }
                                            ]
                                        }
                                    },
                                    {
                                        'sib6': {
                                            't-ReselectionUTRA': 3
                                        }
                                    },
                                    {
                                        'sib7': {
                                            't-ReselectionGERAN': 3
                                        }
                                    },
                                    {
                                        'sib8': {
                                            'parameters1XRTT': {
                                                'longCodeState1XRTT': (b'\x01\x23\x45\x67\x89\x00', 42)
                                            }
                                        }
                                    },
                                    {
                                        'sib9': {
                                            'hnb-Name': b'\x34'
                                        }
                                    },
                                    {
                                        'sib10': {
                                            'messageIdentifier': (b'\x23\x34', 16),
                                            'serialNumber': (b'\x12\x34', 16),
                                            'warningType': b'\x32\x12'
                                        }
                                    },
                                    {
                                        'sib11': {
                                            'messageIdentifier': (b'\x67\x88', 16),
                                            'serialNumber': (b'\x54\x35', 16),
                                            'warningMessageSegmentType': 'notLastSegment',
                                            'warningMessageSegmentNumber': 19,
                                            'warningMessageSegment': b'\x12'
                                        }
                                    }
                                ]
                            }
                        }
                    }
                }
            }
        }

        encoded = (
            b'\x04\x81\x3f\xbe\x2a\x64\x12\xb2\xf3\x3a\x24\x2a\x80\x02\x02\x9b'
            b'\x29\x8a\x7f\xf8\x24\x00\x00\x11\x00\x24\xe2\x08\x05\x06\xc3\xc4'
            b'\x76\x92\x81\x41\x00\xc0\x00\x00\x0b\x23\xfd\x10\x80\xca\x19\x82'
            b'\x80\x48\xd1\x59\xe2\x43\xa0\x1a\x20\x23\x34\x12\x34\x32\x12\x48'
            b'\xcf\x10\xa8\x6a\x4c\x04\x48'
        )

        self.assert_encode_decode(rrc, 'BCCH-DL-SCH-Message', decoded, encoded)

        # Message 5.
        decoded = {
            'message': {
                'c1': {
                    'counterCheck': {
                        'rrc-TransactionIdentifier': 0,
                        'criticalExtensions': {
                            'criticalExtensionsFuture': {}
                        }
                    }
                }
            }
        }

        encoded = b'\x41'

        self.assert_encode_decode(rrc, 'DL-DCCH-Message', decoded, encoded)

        # Message 6.
        decoded = {
            'message': {
                'c1': {
                    'counterCheck': {
                        'rrc-TransactionIdentifier': 0,
                        'criticalExtensions': {
                            'c1': {
                                'counterCheck-r8': {
                                    'drb-CountMSB-InfoList': [
                                        {
                                            'drb-Identity': 32,
                                            'countMSB-Uplink': 33554431,
                                            'countMSB-Downlink': 33554431
                                        }
                                    ],
                                    'nonCriticalExtension': {}
                                }
                            }
                        }
                    }
                }
            }
        }

        encoded = b'\x40\x21\xff\xff\xff\xff\xff\xff\xfc'

        self.assert_encode_decode(rrc, 'DL-DCCH-Message', decoded, encoded)

        # Message 7.
        decoded = {
            'message': {
                'c1': {
                    'counterCheckResponse': {
                        'rrc-TransactionIdentifier': 0,
                        'criticalExtensions': {
                            'counterCheckResponse-r8': {
                                'drb-CountInfoList': [
                                ],
                                'nonCriticalExtension': {}
                            }
                        }
                    }
                }
            }
        }

        encoded = b'\x50\x80'

        self.assert_encode_decode(rrc, 'UL-DCCH-Message', decoded, encoded)

    def test_lpp_14_3_0(self):
        lpp = asn1tools.compile_dict(deepcopy(LPP_14_3_0), 'uper')

        # Message 1.
        decoded = {
            'transactionID': {
                'initiator': 'targetDevice',
                'transactionNumber': 254
            },
            'endTransaction': True,
            'lpp-MessageBody': {
                'c1': {
                    'provideAssistanceData': {
                        'criticalExtensions': {
                            'c1': {
                                'spare1': None
                            }
                        }
                    }
                }
            }
        }

        encoded = b'\x93\xfd\x1b'

        self.assert_encode_decode(lpp, 'LPP-Message', decoded, encoded)

    def test_all_types(self):
        all_types = asn1tools.compile_files('tests/files/all_types.asn',
                                            'uper')

        datas = [
            ('Boolean', True, b'\x80'),
            ('Boolean', False, b'\x00'),
            ('Integer', 32768, b'\x03\x00\x80\x00'),
            ('Integer', 32767, b'\x02\x7f\xff'),
            ('Integer', 256, b'\x02\x01\x00'),
            ('Integer', 255, b'\x02\x00\xff'),
            ('Integer', 128, b'\x02\x00\x80'),
            ('Integer', 127, b'\x01\x7f'),
            ('Integer', 1, b'\x01\x01'),
            ('Integer', 0, b'\x01\x00'),
            ('Integer', -1, b'\x01\xff'),
            ('Integer', -128, b'\x01\x80'),
            ('Integer', -129, b'\x02\xff\x7f'),
            ('Integer', -256, b'\x02\xff\x00'),
            ('Integer', -32768, b'\x02\x80\x00'),
            ('Integer', -32769, b'\x03\xff\x7f\xff'),
            ('Bitstring', (b'\x40', 4), b'\x04\x40'),
            ('Bitstring2', (b'\x12\x80', 9), b'\x12\x80'),
            ('Bitstring3', (b'\x34', 6), b'\x4d'),
            ('Octetstring', b'\x00', b'\x01\x00'),
            ('Octetstring', 500 * b'\x00', b'\x81\xf4' + 500 * b'\x00'),
            ('Octetstring2', b'\xab\xcd', b'\xab\xcd'),
            ('Octetstring3', b'\xab\xcd\xef', b'\xab\xcd\xef'),
            ('Octetstring4', b'\x89\xab\xcd\xef', b'\x31\x35\x79\xbd\xe0'),
            ('Enumerated', 'one', b'\x00'),
            ('Enumerated2', 'two', b'\x40'),
            ('Sequence', {}, b''),
            ('Sequence2', {'a': 0}, b'\x00'),
            ('Sequence2', {'a': 1}, b'\x80\x80\x80'),
            ('Sequence3', {'a': True}, b'\x40'),
            ('Sequence4', {'a': True}, b'\x40'),
            ('Ia5string', 'bar', b'\x03\xc5\x87\x90'),
            ('Utf8string', u'bar', b'\x03\x62\x61\x72'),
            ('Utf8string', u'a\u1010c', b'\x05\x61\xe1\x80\x90\x63')
        ]

        for type_name, decoded, encoded in datas:
            self.assert_encode_decode(all_types, type_name, decoded, encoded)

    def test_encode_all_types(self):
        all_types = asn1tools.compile_files('tests/files/all_types.asn',
                                            'uper')

        self.assertEqual(all_types.encode('Sequence2', {}), b'\x00')

        with self.assertRaises(AssertionError):
            self.assertEqual(all_types.encode('Sequence4', {'a': 1, 'b': True}),
                             b'\xc0\x40\x60\x00')

        self.assertEqual(all_types.encode('Sequence5', {'a': True}), b'\x40')

        with self.assertRaises(AssertionError):
            self.assertEqual(all_types.encode('Sequence5', {'a': True, 'b': True}),
                             b'\xc0\x40\x60\x00')

        with self.assertRaises(AssertionError):
            self.assertEqual(all_types.encode('Sequence6', {'a': True, 'c': True}),
                             b'\x60')

        with self.assertRaises(AssertionError):
            self.assertEqual(all_types.encode('Sequence6',
                                              {'a': True, 'b': True, 'c': True}),
                             b'\xe0\x20\x30\x00')

        with self.assertRaises(AssertionError):
            self.assertEqual(all_types.encode('Sequence7', {'a': True, 'd': True}),
                             b'\x60')

        with self.assertRaises(AssertionError):
            self.assertEqual(all_types.encode('Sequence7',
                                              {'a': True, 'b': True, 'd': True}),
                             b'\xe0\x60\x18\x00')

        with self.assertRaises(AssertionError):
            self.assertEqual(
                all_types.encode('Sequence7',
                                 {'a': True, 'b': True, 'c': True, 'd': True}),
                b'\xe0\x70\x18\x00\x18\x00')

        with self.assertRaises(NotImplementedError):
            all_types.encode('Sequence12', {'a': [{'a': []}]})

        with self.assertRaises(NotImplementedError):
            all_types.encode('Real', 0.0)

        with self.assertRaises(NotImplementedError):
            all_types.decode('Real', b'\x00')

        with self.assertRaises(NotImplementedError):
            all_types.encode('Numericstring', '')

        with self.assertRaises(NotImplementedError):
            all_types.decode('Numericstring', b'\x00')

        with self.assertRaises(NotImplementedError):
            all_types.encode('SetOf', [])

        with self.assertRaises(NotImplementedError):
            all_types.decode('SetOf', b'\x00')

    def test_decode_all_types(self):
        all_types = asn1tools.compile_files('tests/files/all_types.asn',
                                            'uper')

        with self.assertRaises(NotImplementedError):
            all_types.decode('Sequence12', b'\x80\x80')

    def test_repr_all_types(self):
        all_types = asn1tools.compile_files('tests/files/all_types.asn',
                                            'uper')

        self.assertEqual(repr(all_types.types['Boolean']), 'Boolean(Boolean)')
        self.assertEqual(repr(all_types.types['Integer']), 'Integer(Integer)')
        self.assertEqual(repr(all_types.types['Real']), 'Real(Real)')
        self.assertEqual(repr(all_types.types['Bitstring']), 'BitString(Bitstring)')
        self.assertEqual(repr(all_types.types['Octetstring']), 'OctetString(Octetstring)')
        self.assertEqual(repr(all_types.types['Null']), 'Null(Null)')
        self.assertEqual(repr(all_types.types['Objectidentifier']),
                         'ObjectIdentifier(Objectidentifier)')
        self.assertEqual(repr(all_types.types['Enumerated']), 'Enumerated(Enumerated)')
        self.assertEqual(repr(all_types.types['Utf8string']), 'UTF8String(Utf8string)')
        self.assertEqual(repr(all_types.types['Sequence']), 'Sequence(Sequence, [])')
        self.assertEqual(repr(all_types.types['Set']), 'Set(Set, [])')
        self.assertEqual(repr(all_types.types['Sequence2']),
                         'Sequence(Sequence2, [Integer(a)])')
        self.assertEqual(repr(all_types.types['Set2']), 'Set(Set2, [Integer(a)])')
        self.assertEqual(repr(all_types.types['Numericstring']),
                         'NumericString(Numericstring)')
        self.assertEqual(repr(all_types.types['Printablestring']),
                         'PrintableString(Printablestring)')
        self.assertEqual(repr(all_types.types['Ia5string']), 'IA5String(Ia5string)')
        self.assertEqual(repr(all_types.types['Universalstring']),
                         'UniversalString(Universalstring)')
        self.assertEqual(repr(all_types.types['Visiblestring']),
                         'VisibleString(Visiblestring)')
        self.assertEqual(repr(all_types.types['Generalstring']),
                         'GeneralString(Generalstring)')
        self.assertEqual(repr(all_types.types['Bmpstring']),
                         'BMPString(Bmpstring)')
        self.assertEqual(repr(all_types.types['Teletexstring']),
                         'TeletexString(Teletexstring)')
        self.assertEqual(repr(all_types.types['Utctime']), 'UTCTime(Utctime)')
        self.assertEqual(repr(all_types.types['SequenceOf']),
                         'SequenceOf(SequenceOf, Integer())')
        self.assertEqual(repr(all_types.types['SetOf']), 'SetOf(SetOf, Integer())')

    def test_all_types_automatic_tags(self):
        all_types = asn1tools.compile_files(
            'tests/files/all_types_automatic_tags.asn', 'uper')

        datas = [
            ('Sequence3',
             {'a': 1, 'c': 2,'d': True},
             b'\x00\x80\x80\x81\x40')
        ]

        for type_name, decoded, encoded in datas:
            self.assert_encode_decode(all_types, type_name, decoded, encoded)

    def test_utf8_string(self):
        foo = asn1tools.compile_string(
            "Foo DEFINITIONS AUTOMATIC TAGS ::= "
            "BEGIN "
            "S ::= SEQUENCE { "
            "    a BOOLEAN, "
            "    b UTF8String, "
            "    c UTF8String OPTIONAL"
            "} "
            "END",
            'uper')

        datas = [
            ('S', {'a': True, 'b': u''}, b'\x40\x00'),
            ('S',
             {'a': True, 'b': u'1', 'c': u'foo'},
             b'\xc0\x4c\x40\xd9\x9b\xdb\xc0'),
            ('S',
             {'a': True, 'b': 300 * u'1'},
             b'\x60\x4b\x0c' + 299 * b'\x4c' + b'\x40')
        ]

        for type_name, decoded, encoded in datas:
            self.assert_encode_decode(foo, type_name, decoded, encoded)

        with self.assertRaises(NotImplementedError):
            foo.encode('S', {'a': False, 'b': 16384 * u'0'})

        with self.assertRaises(NotImplementedError):
            foo.decode('S', b'\x70\x00\x00\x00')


if __name__ == '__main__':
    unittest.main()



import unittest
from .utils import Asn1ToolsBaseTest
import asn1tools
import sys
from copy import deepcopy

sys.path.append('tests/files/3gpp')

from rrc_8_6_0 import EXPECTED as RRC_8_6_0


class Asn1ToolsXerTest(Asn1ToolsBaseTest):

    maxDiff = None

    def test_foo(self):
        foo = asn1tools.compile_files(['tests/files/foo.asn'], 'xer')

        self.assertEqual(len(foo.types), 2)
        self.assertTrue(foo.types['Question'] is not None)
        self.assertTrue(foo.types['Answer'] is not None)
        self.assertEqual(len(foo.modules), 1)
        self.assertTrue(foo.modules['Foo'] is not None)

        # Question.
        decoded_message = {'id': 1, 'question': 'Is 1+1=3?'}
        encoded_message = (
            b'<Question><id>1</id><question>Is 1+1=3?</question></Question>'
        )

        encoded = foo.encode('Question', decoded_message)
        self.assertEqual(encoded, encoded_message)
        decoded = foo.decode('Question', encoded)
        self.assertEqual(decoded, decoded_message)

        # Answer.
        decoded_message = {'id': 1, 'answer': False}
        encoded_message = (
            b'<Answer><id>1</id><answer><false /></answer></Answer>'
        )

        encoded = foo.encode('Answer', decoded_message)
        self.assertEqual(encoded, encoded_message)
        decoded = foo.decode('Answer', encoded)
        self.assertEqual(decoded, decoded_message)

        # Encode a question with missing field 'id'.
        with self.assertRaises(asn1tools.EncodeError) as cm:
            encoded = foo.encode('Question', {'question': 'Is 1+1=3?'})

        self.assertEqual(
            str(cm.exception),
            "Sequence member 'id' not found in {'question': 'Is 1+1=3?'}.")

    def test_decode_length(self):
        foo = asn1tools.compile_files('tests/files/foo.asn', 'xer')

        with self.assertRaises(asn1tools.DecodeError) as cm:
            foo.decode_length(b'')

        self.assertEqual(str(cm.exception),
                         ': Decode length not supported for this codec.')

    def test_rrc_8_6_0(self):
        rrc = asn1tools.compile_dict(deepcopy(RRC_8_6_0), 'xer')

        # Message 1.
        decoded = {
            'message': {
                'c1' : {
                    'paging': {
                        'systemInfoModification': 'true',
                        'nonCriticalExtension': {
                        }
                    }
                }
            }
        }

        encoded = (
            b'<PCCH-Message><message><c1><paging><systemInfoModification><tr'
            b'ue /></systemInfoModification><nonCriticalExtension /></paging><'
            b'/c1></message></PCCH-Message>'
        )

        self.assert_encode_decode(rrc, 'PCCH-Message', decoded, encoded)

        return

        # Message 2.
        decoded = {
            'message': {
                'dl-Bandwidth': 'n6',
                'phich-Config': {
                    'phich-Duration': 'normal',
                    'phich-Resource': 'half'
                },
                'systemFrameNumber': (b'\x12', 8),
                'spare': (b'\x34\x40', 10)
            }
        }

        encoded = (
            b''
        )

        self.assert_encode_decode(rrc, 'BCCH-BCH-Message', decoded, encoded)

        # Message 3.
        decoded = {
            'message': {
                'c1': {
                    'systemInformation': {
                        'criticalExtensions': {
                            'systemInformation-r8': {
                                'sib-TypeAndInfo': [
                                    {
                                        'sib2': {
                                            'ac-BarringInfo': {
                                                'ac-BarringForEmergency': True,
                                                'ac-BarringForMO-Data': {
                                                    'ac-BarringFactor': 'p95',
                                                    'ac-BarringTime': 's128',
                                                    'ac-BarringForSpecialAC': (b'\xf0', 5)
                                                }
                                            },
                                            'radioResourceConfigCommon': {
                                                'rach-ConfigCommon': {
                                                    'preambleInfo': {
                                                        'numberOfRA-Preambles': 'n24',
                                                        'preamblesGroupAConfig': {
                                                            'sizeOfRA-PreamblesGroupA': 'n28',
                                                            'messageSizeGroupA': 'b144',
                                                            'messagePowerOffsetGroupB': 'minusinfinity'
                                                        }
                                                    },
                                                    'powerRampingParameters': {
                                                        'powerRampingStep': 'dB0',
                                                        'preambleInitialReceivedTargetPower': 'dBm-102'
                                                    },
                                                    'ra-SupervisionInfo': {
                                                        'preambleTransMax': 'n8',
                                                        'ra-ResponseWindowSize': 'sf6',
                                                        'mac-ContentionResolutionTimer': 'sf48'
                                                    },
                                                    'maxHARQ-Msg3Tx': 8
                                                },
                                                'bcch-Config': {
                                                    'modificationPeriodCoeff': 'n2'
                                                },
                                                'pcch-Config': {
                                                    'defaultPagingCycle': 'rf256',
                                                    'nB': 'twoT'
                                                },
                                                'prach-Config': {
                                                    'rootSequenceIndex': 836,
                                                    'prach-ConfigInfo': {
                                                        'prach-ConfigIndex': 33,
                                                        'highSpeedFlag': False,
                                                        'zeroCorrelationZoneConfig': 10,
                                                        'prach-FreqOffset': 64
                                                    }
                                                },
                                                'pdsch-ConfigCommon': {
                                                    'referenceSignalPower': -60,
                                                    'p-b': 2
                                                },
                                                'pusch-ConfigCommon': {
                                                    'pusch-ConfigBasic': {
                                                        'n-SB': 1,
                                                        'hoppingMode': 'interSubFrame',
                                                        'pusch-HoppingOffset': 10,
                                                        'enable64QAM': False
                                                    },
                                                    'ul-ReferenceSignalsPUSCH': {
                                                        'groupHoppingEnabled': True,
                                                        'groupAssignmentPUSCH': 22,
                                                        'sequenceHoppingEnabled': False,
                                                        'cyclicShift': 5
                                                    }
                                                },
                                                'pucch-ConfigCommon': {
                                                    'deltaPUCCH-Shift': 'ds1',
                                                    'nRB-CQI': 98,
                                                    'nCS-AN': 4,
                                                    'n1PUCCH-AN': 2047
                                                },
                                                'soundingRS-UL-ConfigCommon': {
                                                    'setup': {
                                                        'srs-BandwidthConfig': 'bw0',
                                                        'srs-SubframeConfig': 'sc4',
                                                        'ackNackSRS-SimultaneousTransmission': True
                                                    }
                                                },
                                                'uplinkPowerControlCommon': {
                                                    'p0-NominalPUSCH': -126,
                                                    'alpha': 'al0',
                                                    'p0-NominalPUCCH': -127,
                                                    'deltaFList-PUCCH': {
                                                        'deltaF-PUCCH-Format1': 'deltaF-2',
                                                        'deltaF-PUCCH-Format1b': 'deltaF1',
                                                        'deltaF-PUCCH-Format2': 'deltaF0',
                                                        'deltaF-PUCCH-Format2a': 'deltaF-2',
                                                        'deltaF-PUCCH-Format2b': 'deltaF0'
                                                    },
                                                    'deltaPreambleMsg3': -1
                                                },
                                                'ul-CyclicPrefixLength': 'len1'
                                            },
                                            'ue-TimersAndConstants': {
                                                't300': 'ms100',
                                                't301': 'ms200',
                                                't310': 'ms50',
                                                'n310': 'n2',
                                                't311': 'ms30000',
                                                'n311': 'n2'
                                            },
                                            'freqInfo': {
                                                'additionalSpectrumEmission': 3
                                            },
                                            'timeAlignmentTimerCommon': 'sf500'
                                        }
                                    },
                                    {
                                        'sib3': {
                                            'cellReselectionInfoCommon': {
                                                'q-Hyst': 'dB0',
                                                'speedStateReselectionPars': {
                                                    'mobilityStateParameters': {
                                                        't-Evaluation': 's180',
                                                        't-HystNormal': 's180',
                                                        'n-CellChangeMedium': 1,
                                                        'n-CellChangeHigh': 16
                                                    },
                                                    'q-HystSF': {
                                                        'sf-Medium': 'dB-6',
                                                        'sf-High': 'dB-4'
                                                    }
                                                }
                                            },
                                            'cellReselectionServingFreqInfo': {
                                                'threshServingLow': 7,
                                                'cellReselectionPriority': 3
                                            },
                                            'intraFreqCellReselectionInfo': {
                                                'q-RxLevMin': -33,
                                                's-IntraSearch': 0,
                                                'presenceAntennaPort1': False,
                                                'neighCellConfig': (b'\x80', 2),
                                                't-ReselectionEUTRA': 4
                                            }
                                        }
                                    },
                                    {
                                        'sib4': {
                                        }
                                    },
                                    {
                                        'sib5': {
                                            'interFreqCarrierFreqList': [
                                                {
                                                    'dl-CarrierFreq': 1,
                                                    'q-RxLevMin': -45,
                                                    't-ReselectionEUTRA': 0,
                                                    'threshX-High': 31,
                                                    'threshX-Low': 29 ,
                                                    'allowedMeasBandwidth': 'mbw6',
                                                    'presenceAntennaPort1': True,
                                                    'q-OffsetFreq': 'dB0',
                                                    'neighCellConfig': (b'\x00', 2)
                                                }
                                            ]
                                        }
                                    },
                                    {
                                        'sib6': {
                                            't-ReselectionUTRA': 3
                                        }
                                    },
                                    {
                                        'sib7': {
                                            't-ReselectionGERAN': 3
                                        }
                                    },
                                    {
                                        'sib8': {
                                            'parameters1XRTT': {
                                                'longCodeState1XRTT': (b'\x01\x23\x45\x67\x89\x00', 42)
                                            }
                                        }
                                    },
                                    {
                                        'sib9': {
                                            'hnb-Name': b'\x34'
                                        }
                                    },
                                    {
                                        'sib10': {
                                            'messageIdentifier': (b'\x23\x34', 16),
                                            'serialNumber': (b'\x12\x34', 16),
                                            'warningType': b'\x32\x12'
                                        }
                                    },
                                    {
                                        'sib11': {
                                            'messageIdentifier': (b'\x67\x88', 16),
                                            'serialNumber': (b'\x54\x35', 16),
                                            'warningMessageSegmentType': 'notLastSegment',
                                            'warningMessageSegmentNumber': 19,
                                            'warningMessageSegment': b'\x12'
                                        }
                                    }
                                ]
                            }
                        }
                    }
                }
            }
        }

        encoded = (
            b''
        )

        self.assert_encode_decode(rrc, 'BCCH-DL-SCH-Message', decoded, encoded)

    def test_all_types(self):
        all_types = asn1tools.compile_files('tests/files/all_types.asn', 'xer')

        datas = [
            ('Boolean',        True, b'<Boolean><true /></Boolean>'),
            ('Integer',         127, b'<Integer>127</Integer>'),
            ('Integer',           0, b'<Integer>0</Integer>'),
            ('Integer',        -128, b'<Integer>-128</Integer>'),
            ('Enumerated',    'one', b'<Enumerated><one /></Enumerated>'),
            ('Sequence',         {}, b'<Sequence />'),
            ('Set',              {}, b'<Set />'),
            ('Sequence2',  {'a': 1}, b'<Sequence2><a>1</a></Sequence2>'),
            ('Set2',       {'a': 2}, b'<Set2><a>2</a></Set2>'),
            ('Ia5string',     'bar', b'<Ia5string>bar</Ia5string>'),
            ('SequenceOf',       [], b'<SequenceOf />'),
            ('SetOf',            [], b'<SetOf />')
        ]

        for type_name, decoded, encoded in datas:
            self.assert_encode_decode(all_types, type_name, decoded, encoded)

        with self.assertRaises(NotImplementedError):
            all_types.encode('Sequence12', {'a': [{'a': []}]})

        with self.assertRaises(NotImplementedError):
            all_types.decode('Sequence12',
                             b'<Sequence12><a><Sequence12/></a></Sequence12>')

    def test_repr_all_types(self):
        all_types = asn1tools.compile_files('tests/files/all_types.asn',
                                            'xer')

        self.assertEqual(repr(all_types.types['Boolean']), 'Boolean(Boolean)')
        self.assertEqual(repr(all_types.types['Integer']), 'Integer(Integer)')
        self.assertEqual(repr(all_types.types['Real']), 'Real(Real)')
        self.assertEqual(repr(all_types.types['Bitstring']), 'BitString(Bitstring)')
        self.assertEqual(repr(all_types.types['Octetstring']), 'OctetString(Octetstring)')
        self.assertEqual(repr(all_types.types['Null']), 'Null(Null)')
        self.assertEqual(repr(all_types.types['Objectidentifier']),
                         'ObjectIdentifier(Objectidentifier)')
        self.assertEqual(repr(all_types.types['Enumerated']), 'Enumerated(Enumerated)')
        self.assertEqual(repr(all_types.types['Utf8string']), 'UTF8String(Utf8string)')
        self.assertEqual(repr(all_types.types['Sequence']), 'Sequence(Sequence, [])')
        self.assertEqual(repr(all_types.types['Set']), 'Set(Set, [])')
        self.assertEqual(repr(all_types.types['Sequence2']),
                         'Sequence(Sequence2, [Integer(a)])')
        self.assertEqual(repr(all_types.types['Set2']), 'Set(Set2, [Integer(a)])')
        self.assertEqual(repr(all_types.types['Numericstring']),
                         'NumericString(Numericstring)')
        self.assertEqual(repr(all_types.types['Printablestring']),
                         'PrintableString(Printablestring)')
        self.assertEqual(repr(all_types.types['Ia5string']), 'IA5String(Ia5string)')
        self.assertEqual(repr(all_types.types['Universalstring']),
                         'UniversalString(Universalstring)')
        self.assertEqual(repr(all_types.types['Visiblestring']),
                         'VisibleString(Visiblestring)')
        self.assertEqual(repr(all_types.types['Generalstring']),
                         'GeneralString(Generalstring)')
        self.assertEqual(repr(all_types.types['Bmpstring']),
                         'BMPString(Bmpstring)')
        self.assertEqual(repr(all_types.types['Teletexstring']),
                         'TeletexString(Teletexstring)')
        self.assertEqual(repr(all_types.types['Utctime']), 'UTCTime(Utctime)')
        self.assertEqual(repr(all_types.types['SequenceOf']),
                         'SequenceOf(SequenceOf, Integer())')
        self.assertEqual(repr(all_types.types['SetOf']), 'SetOf(SetOf, Integer())')

    def test_all_types_automatic_tags(self):
        all_types = asn1tools.compile_files(
            'tests/files/all_types_automatic_tags.asn', 'xer')

        datas = [
            ('Sequence3',
             {'a': 1, 'c': 2,'d': True},
             b'<Sequence3><a>1</a><c>2</c><d><true /></d></Sequence3>')
        ]

        for type_name, decoded, encoded in datas:
            self.assert_encode_decode(all_types, type_name, decoded, encoded)


if __name__ == '__main__':
    unittest.main()



"""Convert ASN.1 specifications to Python data structures.

"""

import logging
import re

from pyparsing import Literal
from pyparsing import Keyword
from pyparsing import Word
from pyparsing import ZeroOrMore
from pyparsing import Regex
from pyparsing import printables
from pyparsing import delimitedList
from pyparsing import Group
from pyparsing import Optional
from pyparsing import Forward
from pyparsing import StringEnd
from pyparsing import OneOrMore
from pyparsing import nums
from pyparsing import Suppress
from pyparsing import ParseException
from pyparsing import ParseSyntaxException
from pyparsing import NotAny
from pyparsing import NoMatch
from pyparsing import QuotedString
from pyparsing import Combine
from pyparsing import ParseResults


LOGGER = logging.getLogger(__name__)


class ParseError(Exception):
    pass


class InternalParserError(Exception):
    pass


class Tokens(object):

    def __init__(self, tag, tokens):
        self.tag = tag
        self.tokens = tokens

    def __getitem__(self, index):
        return self.tokens[index]

    def __len__(self):
        return len(self.tokens)

    def __iter__(self):
        for token in self.tokens:
            yield token

    def __bool__(self):
        return len(self.tokens) > 0

    def __eq__(self, other):
        return other == self.tag

    def __repr__(self):
        return "Tokens(tag='{}', tokens='{}')".format(self.tag,
                                                      self.tokens)


class Tag(Group):

    def __init__(self, tag, expr):
        super(Tag, self).__init__(expr)
        self.tag = tag

    def postParse(self, instring, loc, tokenlist):
        return Tokens(self.tag, tokenlist.asList())


def merge_dicts(dicts):
    return {k: v for d in dicts for k, v in d.items()}


def convert_integer(_s, _l, tokens):
    try:
        return int(tokens[0])
    except:
        return tokens


def convert_real_number(_s, _l, tokens):
    if '.' not in tokens[0]:
        tokens = int(tokens[0])

    return tokens


def convert_number(token):
    if isinstance(token, list):
        token = token[0]

    try:
        return int(token)
    except (ValueError, TypeError):
        return token


def convert_size(tokens):
    if len(tokens) == 0:
        return None

    tokens = tokens[0]

    if tokens[0] == 'SIZE':
        values = []

        for item_tokens in tokens[1].asList():
            if '..' in item_tokens:
                value = (convert_number(item_tokens[0]),
                         convert_number(item_tokens[2]))
            else:
                value = convert_number(item_tokens[0])

            values.append(value)

        return values
    elif isinstance(tokens[0], dict):
        return tokens[0]['size']


def convert_table(tokens):
    tokens = tokens[0]

    try:
        if isinstance(tokens[1][0][0], list):
            defined_object_set = tokens[1][0][0][0]
        else:
            defined_object_set = tokens[1][0][0]
    except IndexError:
        return None

    try:
        component_ids = tokens[4]
    except IndexError:
        return defined_object_set

    return [defined_object_set, component_ids]


def convert_enum_values(tokens):
    number = 0
    values = {}

    for token in tokens:
        if len(token) == 2:
            number = int(token[1])

        values[number] = token[0]
        number += 1

    return values


def convert_tag(tokens):
    if len(tokens) > 0:
        if len(tokens[0]) == 1:
            tag = {
                'number': int(tokens[0][0])
            }
        else:
            tag = {
                'number': convert_number(tokens[0][1]),
                'class': tokens[0][0]
            }

        if tokens[1]:
            tag['kind'] = tokens[1][0] if tokens[1] else None

        return tag


def convert_value_range(_s, _l, tokens):
    tokens = tokens.asList()
    minimum = tokens[0]

    if isinstance(minimum, list):
        minimum = minimum[0]

    maximum = tokens[1]

    if isinstance(maximum, list):
        maximum = maximum[0]

    return (minimum, maximum)


def convert_size_constraint(_s, _l, tokens):
    tokens = tokens.asList()[1]
    values = []

    for item_tokens in tokens:
        if '..' in item_tokens:
            value = (convert_number(item_tokens[0]),
                     convert_number(item_tokens[2]))
        else:
            value = convert_number(item_tokens[0])

        values.append(value)

    return {'size': values}


def convert_permitted_alphabet(_s, _l, tokens):
    tokens = tokens.asList()

    return {'from': tokens[1]}


def convert_constraint(_s, _l, tokens):
    tokens = tokens.asList()
    # print('constraint:', tokens)


def convert_members(tokens):
    members = []

    for member_tokens in tokens:
        if member_tokens in [['...'], '...']:
            members.append('...')
            continue

        if member_tokens[0] == 'COMPONENTS OF':
            members.append({
                'components-of': member_tokens[1][0]['type']
            })
            continue

        if len(member_tokens) == 2:
            member_tokens, qualifiers = member_tokens
            qualifiers = qualifiers.asList()
        else:
            qualifiers = []

        member = convert_type(member_tokens[2])
        member['name'] = member_tokens[0]

        if 'OPTIONAL' in qualifiers:
            member['optional'] = True

        if 'DEFAULT' in qualifiers:
            if len(qualifiers[1]) == 0:
                value = []
            else:
                value = convert_value(qualifiers[1], member['type'])

            member['default'] = value

        tag = convert_tag(member_tokens[1])

        if tag:
            member['tag'] = tag

        members.append(member)

    return members


def convert_sequence_type(_s, _l, tokens):
    return {
        'type': 'SEQUENCE',
        'members': convert_members(tokens[2])
    }


def convert_sequence_of_type(_s, _l, tokens):
    converted_type = {
        'type': 'SEQUENCE OF',
        'element': convert_type(tokens[4]),
    }

    size = convert_size(tokens[1])

    if size:
        converted_type['size'] = size

    tag = convert_tag(tokens[3])

    if tag:
        converted_type['element']['tag'] = tag

    return converted_type


def convert_set_type(_s, _l, tokens):
    return {
        'type': 'SET',
        'members': convert_members(tokens[2])
    }


def convert_set_of_type(_s, _l, tokens):
    converted_type = {
        'type': 'SET OF',
        'element': convert_type(tokens[4])
    }

    size = convert_size(tokens[1])

    if size:
        converted_type['size'] = size

    tag = convert_tag(tokens[3])

    if tag:
        converted_type['element']['tag'] = tag

    return converted_type


def convert_choice_type(_s, _l, tokens):
    return {
        'type': 'CHOICE',
        'members': convert_members(tokens[2])
    }


def convert_defined_type(_s, _l, tokens):
    return {
        'type': tokens[0]
    }


def convert_integer_type(_s, _l, _tokens):
    return {'type': 'INTEGER'}


def convert_real_type(_s, _l, _tokens):
    return {'type': 'REAL'}


def convert_enumerated_type(_s, _l, tokens):
    return {
        'type': 'ENUMERATED',
        'values': convert_enum_values(tokens[2])
    }


def convert_keyword_type(_s, _l, tokens):
    return {
        'type': tokens[0]
    }


def convert_print(_s, _l, tokens):
    print('convert_print', tokens)


def convert_object_identifier_type(_s, _l, _tokens):
    return {
        'type': 'OBJECT IDENTIFIER'
    }


def convert_bit_string_type(_s, _l, _tokens):
    return {
        'type': 'BIT STRING'
    }


def convert_octet_string_type(_s, _l, _tokens):
    return {
        'type': 'OCTET STRING'
    }


def convert_ia5_string_type(_s, _l, _tokens):
    return {
        'type': 'IA5String'
    }


def convert_any_defined_by_type(_s, _l, tokens):
    return {
        'type': 'ANY DEFINED BY',
        'value': tokens[1],
        'choices': {}
    }


def convert_null_type(_s, _l, _tokens):
    return {
        'type': 'NULL'
    }


def convert_boolean_type(_s, _l, _tokens):
    return {
        'type': 'BOOLEAN'
    }


def convert_type(tokens):
    converted_type, constraint = tokens

    if converted_type['type'] in ['INTEGER', 'REAL', 'IA5String']:
        restricted_to = []

        for constraint_tokens in constraint:
            if isinstance(constraint_tokens, ParseResults):
                constraint_tokens = constraint_tokens.asList()

            if len(constraint_tokens) != 1:
                continue

            if isinstance(constraint_tokens[0], dict):
                continue

            restricted_to.append(convert_number(constraint_tokens[0]))

        if restricted_to:
            converted_type['restricted-to'] = restricted_to

    if converted_type['type'] in ['BIT STRING', 'OCTET STRING']:
        size = convert_size(constraint)

        if size:
            converted_type['size'] = size

    if '&' in converted_type['type']:
        converted_type['table'] = convert_table(tokens.asList()[1:])

    return converted_type


def convert_bstring(_s, _l, tokens):
    return '0b' + re.sub(r"[\sB']", '', tokens[0])


def convert_hstring(_s, _l, tokens):
    return '0x' + re.sub(r"[\sH']", '', tokens[0])


def convert_bit_string_value(tokens):
    value = tokens[0]

    if value == 'IdentifierList':
        value = value[:]
    elif isinstance(value, str):
        value = value
    else:
        value = None

    return value


def convert_value(tokens, type_=None):
    if type_ == 'INTEGER':
        value = int(tokens[0])
    elif type_ == 'OBJECT IDENTIFIER':
        value = []

        for value_tokens in tokens:
            if len(value_tokens) == 2:
                value.append((value_tokens[0], int(value_tokens[1])))
            else:
                value.append(convert_number(value_tokens[0]))
    elif type_ == 'BOOLEAN':
        value = tokens[0]
    elif tokens[0] == 'BitStringValue':
        value = convert_bit_string_value(tokens[0])
    elif isinstance(tokens[0], str):
        value = convert_number(tokens[0])
    elif isinstance(tokens[0], int):
        value = tokens[0]
    else:
        value = None

    return value


def convert_parameterized_object_set_assignment(_s, _l, tokens):
    members = []

    try:
        for member_tokens in tokens[4].asList():
            if len(member_tokens[0]) == 1:
                member = member_tokens[0][0]
            else:
                member = {}

                for item_tokens in member_tokens[0]:
                    name = item_tokens[0]
                    value = item_tokens[1][0]

                    if isinstance(value, Tokens):
                        value = value[0]
                    member[name] = convert_number(value)

            members.append(member)
    except IndexError:
        pass

    converted_type = {
        'class': tokens[1],
        'members': members
    }

    return ('parameterized-object-set-assignment',
            tokens[0],
            converted_type)


def convert_parameterized_object_assignment(_s, _l, tokens):
    type_ = tokens[1]

    converted_type = {
        'type': type_,
        'value': None
    }

    return ('parameterized-object-assignment',
            tokens[0],
            converted_type)


def convert_parameterized_object_class_assignment(_s, _l, tokens):
    members = []

    for member in tokens[3]:
        if member[0][1].islower():
            converted_member = member[1][0]

            if isinstance(converted_member, Tokens):
                converted_member = converted_member[0]
        else:
            converted_member = {'type': 'OpenType'}

        converted_member['name'] = member[0]

        members.append(converted_member)

    converted_type = {
        'members': members
    }

    return ('parameterized-object-class-assignment',
            tokens[0],
            converted_type)


def convert_parameterized_type_assignment(_s, _l, tokens):
    tokens = tokens.asList()
    converted_type = convert_type(tokens[3])

    try:
        tag = convert_tag(tokens[2])
    except ValueError:
        tag = None

    if tag:
        converted_type['tag'] = tag

    return ('parameterized-type-assignment',
            tokens[0],
            converted_type)


def convert_parameterized_value_assignment(_s, _l, tokens):
    type_ = tokens[1][0][0]

    if isinstance(type_, Tokens):
        type_ = type_[0]
    elif isinstance(type_, dict):
        type_ = type_['type']

    converted_type = {
        'type': type_,
        'value': convert_value(tokens[2], type_)
    }

    return ('parameterized-value-assignment',
            tokens[0],
            converted_type)


def convert_imports(_s, _l, tokens):
    tokens = tokens.asList()
    imports = {}

    if tokens:
        for from_tokens in tokens:
            from_name = from_tokens[2]
            LOGGER.debug("Converting imports from '%s'.", from_name)
            imports[from_name] = from_tokens[0]

    return {'imports': imports}


def convert_assignment_list(_s, _l, tokens):
    types = {}
    values = {}
    object_classes = {}
    object_sets = {}

    for kind, name, value in tokens:
        if kind == 'parameterized-object-set-assignment':
            if name in object_sets:
                LOGGER.warning("Object set '%s' already defined.", name)

            object_sets[name] = value
        elif kind == 'parameterized-object-assignment':
            if name in values:
                LOGGER.warning("Object '%s' already defined.", name)

            values[name] = value
        elif kind == 'parameterized-object-class-assignment':
            if name in object_classes:
                LOGGER.warning("Object class '%s' already defined.", name)

            object_classes[name] = value
        elif kind == 'parameterized-type-assignment':
            if name in types:
                LOGGER.warning("Type '%s' already defined.", name)

            types[name] = value
        elif kind == 'parameterized-value-assignment':
            if name in values:
                LOGGER.warning("Value '%s' already defined.", name)

            values[name] = value
        else:
            raise InternalParserError(
                'Unrecognized assignment kind {}.'.format(kind))

    return {
        'types': types,
        'values': values,
        'object-classes': object_classes,
        'object-sets': object_sets
    }


def convert_module_body(_s, _l, tokens):
    return merge_dicts(tokens)


def convert_module_definition(_s, _l, tokens):
    tokens = tokens.asList()
    module =  tokens[1][0]
    module['extensibility-implied'] = (tokens[0][3] != [])

    if tokens[0][2]:
        module['tags'] = tokens[0][2][0]

    return {tokens[0][0]: module}


def convert_specification(_s, _l, tokens):
    return merge_dicts(tokens)


def create_grammar():
    """Return the ASN.1 grammar as Pyparsing objects.

    """

    # Keywords.
    SEQUENCE = Keyword('SEQUENCE').setName('SEQUENCE')
    SEQUENCE_OF = Keyword('SEQUENCE OF').setName('SEQUENCE OF')
    CHOICE = Keyword('CHOICE').setName('CHOICE')
    ENUMERATED = Keyword('ENUMERATED').setName('ENUMERATED')
    DEFINITIONS = Keyword('DEFINITIONS').setName('DEFINITIONS')
    BEGIN = Keyword('BEGIN').setName('BEGIN')
    END = Keyword('END').setName('END')
    AUTOMATIC = Keyword('AUTOMATIC').setName('AUTOMATIC')
    TAGS = Keyword('TAGS').setName('TAGS')
    OPTIONAL = Keyword('OPTIONAL').setName('OPTIONAL')
    OF = Keyword('OF').setName('OF')
    SIZE = Keyword('SIZE').setName('SIZE')
    INTEGER = Keyword('INTEGER').setName('INTEGER')
    REAL = Keyword('REAL').setName('REAL')
    BIT_STRING = Keyword('BIT STRING').setName('BIT STRING')
    OCTET_STRING = Keyword('OCTET STRING').setName('OCTET STRING')
    DEFAULT = Keyword('DEFAULT').setName('DEFAULT')
    IMPORTS = Keyword('IMPORTS').setName('IMPORTS')
    EXPORTS = Keyword('EXPORTS').setName('EXPORTS')
    FROM = Keyword('FROM').setName('FROM')
    CONTAINING = Keyword('CONTAINING').setName('CONTAINING')
    ENCODED_BY = Keyword('ENCODED_BY').setName('ENCODED_BY')
    IMPLICIT = Keyword('IMPLICIT').setName('IMPLICIT')
    EXPLICIT = Keyword('EXPLICIT').setName('EXPLICIT')
    OBJECT_IDENTIFIER = Keyword('OBJECT IDENTIFIER').setName('OBJECT IDENTIFIER')
    UNIVERSAL = Keyword('UNIVERSAL').setName('UNIVERSAL')
    APPLICATION = Keyword('APPLICATION').setName('APPLICATION')
    PRIVATE = Keyword('PRIVATE').setName('PRIVATE')
    SET = Keyword('SET').setName('SET')
    SET_OF = Keyword('SET OF').setName('SET OF')
    ANY_DEFINED_BY = Keyword('ANY DEFINED BY').setName('ANY DEFINED BY')
    EXTENSIBILITY_IMPLIED = Keyword('EXTENSIBILITY IMPLIED').setName('EXTENSIBILITY IMPLIED')
    BOOLEAN = Keyword('BOOLEAN').setName('BOOLEAN')
    TRUE = Keyword('TRUE').setName('TRUE')
    FALSE = Keyword('FALSE').setName('FALSE')
    CLASS = Keyword('CLASS').setName('CLASS')
    WITH_SYNTAX = Keyword('WITH SYNTAX').setName('WITH SYNTAX')
    UNIQUE = Keyword('UNIQUE').setName('UNIQUE')
    NULL = Keyword('NULL').setName('NULL')
    WITH_COMPONENT = Keyword('WITH COMPONENT').setName('WITH COMPONENT')
    WITH_COMPONENTS = Keyword('WITH COMPONENTS').setName('WITH COMPONENTS')
    COMPONENTS_OF = Keyword('COMPONENTS OF').setName('COMPONENTS OF')
    PRESENT = Keyword('PRESENT').setName('PRESENT')
    ABSENT = Keyword('ABSENT').setName('ABSENT')
    ALL = Keyword('ALL').setName('ALL')
    MIN = Keyword('MIN').setName('MIN')
    MAX = Keyword('MAX').setName('MAX')
    INCLUDES = Keyword('INCLUDES').setName('INCLUDES')
    PATTERN = Keyword('PATTERN').setName('PATTERN')
    CONSTRAINED_BY = Keyword('CONSTRAINED BY').setName('CONSTRAINED BY')
    UNION = Keyword('UNION').setName('UNION')
    INTERSECTION = Keyword('INTERSECTION').setName('INTERSECTION')
    PLUS_INFINITY = Keyword('PLUS-INFINITY').setName('PLUS-INFINITY')
    MINUS_INFINITY = Keyword('MINUS-INFINITY').setName('MINUS-INFINITY')
    BMPString = Keyword('BMPString').setName('BMPString')
    GeneralString = Keyword('GeneralString').setName('GeneralString')
    GraphicString = Keyword('GraphicString').setName('GraphicString')
    IA5String = Keyword('IA5String').setName('IA5String')
    ISO646String = Keyword('ISO646String').setName('ISO646String')
    NumericString = Keyword('NumericString').setName('NumericString')
    PrintableString = Keyword('PrintableString').setName('PrintableString')
    TeletexString = Keyword('TeletexString').setName('TeletexString')
    UTCTime = Keyword('UTCTime').setName('UTCTime')
    GeneralizedTime = Keyword('GeneralizedTime').setName('GeneralizedTime')
    T61String = Keyword('T61String').setName('T61String')
    UniversalString = Keyword('UniversalString').setName('UniversalString')
    UTF8String = Keyword('UTF8String').setName('UTF8String')
    VideotexString = Keyword('VideotexString').setName('VideotexString')
    VisibleString = Keyword('VisibleString').setName('VisibleString')
    CHARACTER_STRING = Keyword('CHARACTER STRING').setName('CHARACTER STRING')

    # Various literals.
    word = Word(printables, excludeChars=',(){}[].:=;"|').setName('word')
    identifier = Regex(r'[a-z][a-zA-Z0-9-]*').setName('identifier')
    assign = Literal('::=').setName('::=')
    left_parenthesis = Literal('(')
    right_parenthesis = Literal(')')
    left_brace = Literal('{')
    right_brace = Literal('}')
    left_bracket = Literal('[')
    right_bracket = Literal(']')
    left_version_brackets = Literal('[[')
    right_version_brackets = Literal(']]')
    colon = Literal(':')
    semi_colon = Literal(';')
    dot = Literal('.')
    range_separator = Literal('..')
    ellipsis = Literal('...')
    pipe = Literal('|')
    caret = Literal('^')
    comma = Literal(',')
    at = Literal('@')
    integer = Word(nums + '-')
    real_number = Regex(r'[+-]?\d+\.?\d*([eE][+-]?\d+)?')
    bstring = Regex(r"'[01\s]*'B")
    hstring = Regex(r"'[0-9A-F\s]*'H")
    cstring = QuotedString('"')
    number = (Word(nums).setName('number') + ~dot)
    number = Word(printables, excludeChars=',(){}[].:=;"|').setName('number')
    ampersand = Literal('&')
    less_than = Literal('<')

    reserved_words = Regex(r'(END|SEQUENCE|ENUMERATED)(\s|$)')

    # Forward declarations.
    value = Forward()
    type_ = Forward()
    object_ = Forward()
    object_set = Forward()
    primitive_field_name = Forward()
    constraint = Forward()
    element_set_spec = Forward()
    token_or_group_spec = Forward()
    value_reference = Forward().setName('valuereference')
    type_reference = Forward().setName('typereference')
    value_set = Forward().setName('"valueSet" not implemented')
    named_type = Forward()
    root_element_set_spec = Forward()
    defined_object_set = Forward()
    syntax_list = Forward()
    object_from_object = Forward()
    object_set_from_objects = Forward()
    defined_value = Forward().setName('DefinedValue')
    component_type_lists = Forward()
    extension_and_exception = Forward()
    optional_extension_marker = Forward()
    additional_element_set_spec = Forward()
    reference = Forward()
    defined_object_class = Forward()
    defined_type = Forward()
    module_reference = Forward()
    external_type_reference = Forward()
    external_value_reference = Forward()
    simple_defined_type = Forward()
    defined_object = Forward()
    referenced_value = Forward()
    builtin_value = Forward()
    named_value = Forward()
    sequence_value = Forward()
    signed_number = Forward()
    name_and_number_form = Forward()
    number_form = Forward().setName('numberForm')
    definitive_number_form = Forward().setName('definitiveNumberForm')
    version_number = Forward()
    union_mark = Forward()

    value_field_reference = Combine(ampersand + value_reference)
    type_field_reference = Combine(ampersand + type_reference)

    range_ = (word + range_separator + word)

    # ToDo: Remove size and size_paren as they are a workaround for
    # SEQUENCE/SET OF.
    size = Group(SIZE
                 + Group(Suppress(left_parenthesis)
                         + Group(delimitedList(range_ | word, delim=union_mark))
                         + Suppress(right_parenthesis)))

    size_paren = (Suppress(Optional(left_parenthesis))
                  + size
                  + Suppress(Optional(right_parenthesis)))

    class_number = (number | defined_value).setName('ClassNumber')
    tag = Group(Optional(Suppress(left_bracket)
                         - Group(Optional(UNIVERSAL
                                          | APPLICATION
                                          | PRIVATE)
                                 + class_number)
                         - Suppress(right_bracket)
                         + Group(Optional(IMPLICIT | EXPLICIT))))

    any_defined_by_type = (ANY_DEFINED_BY + word)
    any_defined_by_type.setName('ANY DEFINED BY')

    identifier_list = delimitedList(identifier)

    # X.683: 8. Parameterized assignments
    dummy_reference = reference
    dummy_governor = dummy_reference
    governor = (type_ | defined_object_class)
    param_governor = (governor | dummy_governor)
    parameter = (Optional(param_governor + colon) + dummy_reference)
    parameter_list = Suppress(Optional(left_brace
                                       + delimitedList(parameter)
                                       + right_brace))

    # X.683: 9. Referencing parameterized definitions
    actual_parameter = Group(type_
                             | value
                             | value_set
                             | defined_object_class
                             | object_
                             | object_set)
    actual_parameter_list = Group(Suppress(left_brace)
                                  + delimitedList(actual_parameter)
                                  + Suppress(right_brace))
    parameterized_object = (defined_object + actual_parameter_list)
    parameterized_object_set = (defined_object_set + actual_parameter_list)
    parameterized_object_class = (defined_object_class + actual_parameter_list)
    parameterized_value_set_type = (simple_defined_type
                                    + actual_parameter_list)
    simple_defined_value = (external_value_reference
                            | value_reference)
    parameterized_value = (simple_defined_value
                           + actual_parameter_list)
    simple_defined_type <<= (external_type_reference
                             | type_reference)
    parameterized_type = (simple_defined_type
                          + actual_parameter_list)
    parameterized_reference = (reference + Optional(left_brace + right_brace))

    # X.682: 11. Contents constraints
    contents_constraint = ((CONTAINING + type_)
                           | (ENCODED_BY + value)
                           | (CONTAINING + type_ + ENCODED_BY + value))

    # X.682: 10. Table constraints, including component relation constraints
    level = OneOrMore(dot)
    component_id_list = identifier
    at_notation = (Suppress(at)
                   - (component_id_list
                      | Combine(level + component_id_list)))
    component_relation_constraint = (left_brace
                                     + Group(Group(defined_object_set))
                                     + right_brace
                                     + left_brace
                                     - Group(delimitedList(at_notation))
                                     - right_brace)
    component_relation_constraint.setName('"{"')
    simple_table_constraint = object_set
    table_constraint = (component_relation_constraint
                        | simple_table_constraint)

    # X.682: 9. User-defined constants
    user_defined_constraint_parameter = ((governor
                                          + colon
                                          + (value
                                             | value_set
                                             | object_
                                             | object_set))
                                         | type_
                                         | defined_object_class)
    user_defined_constraint = (CONSTRAINED_BY
                               - left_brace
                               - Optional(delimitedList(user_defined_constraint_parameter))
                               - right_brace)
    user_defined_constraint.setName('CONSTRAINED_BY')

    # X.682: 8. General constraint specification
    general_constraint = (user_defined_constraint
                          | table_constraint
                          | contents_constraint)

    # X.681: 7. ASN.1 lexical items
    object_set_reference = type_reference
    value_set_field_reference = NoMatch().setName('"valueSetFieldReference" not implemented')
    object_field_reference = NoMatch().setName('"objectFieldReference" not implemented')
    object_set_field_reference = NoMatch().setName('"objectSetFieldReference" not implemented')
    object_class_reference = (NotAny(reserved_words)
                              + Regex(r'[A-Z][A-Z0-9-]*'))
    object_reference = value_reference

    # X.681: 8. Referencing definitions
    external_object_set_reference = NoMatch().setName('"externalObjectSetReference" not implemented')
    defined_object_set <<= (external_object_set_reference
                            | object_set_reference)
    defined_object <<= NoMatch().setName('"definedObject" not implemented')
    defined_object_class <<= object_class_reference

    # X.681: 9. Information object class definition and assignment
    field_name = primitive_field_name
    primitive_field_name <<= (type_field_reference
                              | value_field_reference
                              | value_set_field_reference
                              | object_field_reference
                              | object_set_field_reference)
    object_set_field_spec = NoMatch().setName('"objectSetFieldSpec" not implemented')
    object_field_spec = NoMatch().setName('"objectFieldSpec" not implemented')
    variable_type_value_set_field_spec = NoMatch().setName(
        '"variableTypeValueSetFieldSpec" not implemented')
    fixed_type_value_set_field_spec = NoMatch().setName('"fixedTypeValueSetFieldSpec" not implemented')
    variable_type_value_field_spec = NoMatch().setName('"variableTypeValueFieldSpec" not implemented')
    fixed_type_value_field_spec = (value_field_reference
                                   + type_
                                   + Optional(UNIQUE)
                                   + Optional(OPTIONAL
                                              | (DEFAULT - value)))
    type_field_spec = (type_field_reference
                       + Optional(OPTIONAL
                                  | (DEFAULT - type_)))
    field_spec = Group(type_field_spec
                       | fixed_type_value_field_spec
                       | variable_type_value_field_spec
                       | fixed_type_value_set_field_spec
                       | variable_type_value_set_field_spec
                       | object_field_spec
                       | object_set_field_spec)
    with_syntax_spec = (WITH_SYNTAX - syntax_list)
    object_class_defn = (CLASS
                         - Suppress(left_brace)
                         - Group(delimitedList(field_spec))
                         - Suppress(right_brace)
                         - Optional(with_syntax_spec))
    object_class = (object_class_defn
                    # | defined_object_class
                    | parameterized_object_class)
    parameterized_object_class_assignment = (object_class_reference
                                             + parameter_list
                                             + assign
                                             + object_class)

    # X.681: 10. Syntax list
    literal = (word | comma)
    required_token = (literal | primitive_field_name)
    optional_group = (left_bracket
                      + OneOrMore(token_or_group_spec)
                      + right_bracket)
    token_or_group_spec <<= (required_token | optional_group)
    syntax_list <<= (left_brace
                     + OneOrMore(token_or_group_spec)
                     + right_brace)

    # X.681: 11. Information object definition and assignment
    setting = (type_ | value | value_set | object_ | object_set | QuotedString('"'))
    field_setting =  Group(primitive_field_name + setting)
    default_syntax = (Suppress(left_brace)
                      + delimitedList(field_setting)
                      + Suppress(right_brace))
    defined_syntax = NoMatch().setName('"definedSyntax" not implemented')
    object_defn = Group(default_syntax | defined_syntax)
    object_ <<= (defined_object
                 | object_defn
                 | object_from_object
                 | parameterized_object)
    parameterized_object_assignment = (object_reference
                                       + parameter_list
                                       + defined_object_class
                                       + Suppress(assign)
                                       + object_)

    # X.681: 12. Information object set definition and assignment
    object_set_elements = (object_
                           | defined_object_set
                           | object_set_from_objects
                           | parameterized_object_set)
    object_set_spec = ((root_element_set_spec
                        + Optional(comma
                                   + ellipsis
                                   + Optional(comma
                                              + additional_element_set_spec)))
                       | (ellipsis + Optional(comma + additional_element_set_spec)))
    object_set <<= (left_brace + Group(object_set_spec) + right_brace)
    object_set.setName('"{"')
    parameterized_object_set_assignment = (object_set_reference
                                           + parameter_list
                                           + defined_object_class
                                           - assign
                                           - object_set)

    # X.681: 13. Associated tables

    # X.681: 14. Notation for the object class field type
    fixed_type_field_val = (builtin_value | referenced_value)
    open_type_field_val = (type_ + colon + value)
    object_class_field_value = (open_type_field_val
                                | fixed_type_field_val)
    object_class_field_type = Combine(defined_object_class
                                      + dot
                                      + field_name)
    object_class_field_type.setName('ObjectClassFieldType')

    # X.681: 15. Information from objects
    object_set_from_objects <<= NoMatch().setName('"objectSetFromObjects" not implemented')
    object_from_object <<= NoMatch().setName('"objectFromObject" not implemented')

    # X.680: 49. The exception identifier
    exception_spec = NoMatch().setName('"exceptionSpec" not implemented')

    # X.680: 47. Subtype elements
    pattern_constraint = (PATTERN + value)
    value_constraint = constraint
    presence_constraint = (PRESENT | ABSENT | OPTIONAL)
    component_constraint = (Optional(value_constraint)
                            + Optional(presence_constraint))
    named_constraint = (identifier + component_constraint)
    type_constraints = delimitedList(named_constraint)
    full_specification = (left_brace + type_constraints + right_brace)
    partial_specification = (left_brace + ellipsis + comma + type_constraints + right_brace)
    single_type_constraint = constraint
    multiple_type_constraints = (full_specification | partial_specification)
    inner_type_constraints = ((WITH_COMPONENT - single_type_constraint)
                              | (WITH_COMPONENTS - multiple_type_constraints))
    permitted_alphabet = (FROM - constraint)
    type_constraint = type_
    size_constraint = (SIZE - Group(constraint))
    upper_end_value = (value | MAX)
    lower_end_value = (value | MIN)
    upper_endpoint = (Optional(less_than) + upper_end_value)
    lower_endpoint = (lower_end_value + Optional(less_than))
    value_range = (((Combine(integer + dot) + Suppress(range_separator))
                    | (integer + Suppress(range_separator))
                    | (lower_endpoint + Suppress(range_separator)))
                   - upper_endpoint)
    contained_subtype = (Optional(INCLUDES) + type_)
    single_value = value
    subtype_elements = (size_constraint
                        | permitted_alphabet
                        | value_range
                        | inner_type_constraints
                        | single_value
                        | pattern_constraint
                        | contained_subtype
                        | type_constraint)

    # X.680: 46. Element set specification
    union_mark <<= (pipe | UNION)
    intersection_mark = (caret | INTERSECTION)
    elements = Group(subtype_elements
                     | object_set_elements
                     | (left_parenthesis + element_set_spec + right_parenthesis))
    unions = delimitedList(elements, delim=(union_mark | intersection_mark))
    element_set_spec <<= unions
    root_element_set_spec <<= element_set_spec
    additional_element_set_spec <<= element_set_spec
    element_set_specs = (root_element_set_spec
                         + Optional(Suppress(comma - ellipsis)
                                    + Optional(Suppress(comma)
                                               - additional_element_set_spec)))

    # X.680: 45. Constrained types
    subtype_constraint = element_set_specs
    constraint_spec = (general_constraint
                       | subtype_constraint)
    constraint_spec.setName('one or more constraints')
    constraint <<= (Suppress(left_parenthesis)
                    - constraint_spec
                    - Suppress(right_parenthesis))

    # X.680: 40. Definition of unrestricted character string types
    unrestricted_character_string_type = CHARACTER_STRING
    unrestricted_character_string_value = NoMatch().setName(
        '"unrestrictedCharacterStringValue" not implemented')

    # X.680: 39. Canonical order of characters

    # X.680: 38. Specification of the ASN.1 module "ASN.1-CHARACTER-MODULE"

    # X.680: 37. Definition of restricted character string types
    group = number
    plane = number
    row = number
    cell = number
    quadruple = (left_brace
                 + group + comma
                 + plane + comma
                 + row + comma +
                 cell
                 + right_brace)
    table_column = number
    table_row = number
    tuple_ = (left_brace + table_column + comma + table_row + right_brace)
    chars_defn = (cstring | quadruple | tuple_ | defined_value)
    charsyms = delimitedList(chars_defn)
    character_string_list = (left_brace + charsyms + right_brace)
    restricted_character_string_value = (cstring
                                         | character_string_list
                                         | quadruple
                                         | tuple_)
    restricted_character_string_type = (BMPString
                                        | GeneralString
                                        | GraphicString
                                        | IA5String
                                        | ISO646String
                                        | NumericString
                                        | PrintableString
                                        | TeletexString
                                        | UTCTime
                                        | GeneralizedTime
                                        | T61String
                                        | UniversalString
                                        | UTF8String
                                        | VideotexString
                                        | VisibleString)

    # X.680: 36. Notation for character string types
    character_string_value = (restricted_character_string_value
                              | unrestricted_character_string_value)
    character_string_type = (restricted_character_string_type
                             | unrestricted_character_string_type)

    # X.680: 35. The character string types

    # X.680: 34. Notation for the external type
    external_value = sequence_value

    # X.680: 33. Notation for embedded-pdv type
    embedded_pdv_value = sequence_value

    # X.680: 32. Notation for relative object identifier type
    relative_oid_components = Group(number_form
                                    | name_and_number_form
                                    | defined_value)
    relative_oid_component_list = OneOrMore(relative_oid_components)
    relative_oid_value = (Suppress(left_brace)
                          + relative_oid_component_list
                          + Suppress(right_brace))

    # X.680: 31. Notation for object identifier type
    name_and_number_form <<= (identifier
                              + Suppress(left_parenthesis)
                              - number_form
                              - Suppress(right_parenthesis))
    number_form <<= (number | defined_value)
    name_form = identifier
    obj_id_components = Group(name_and_number_form
                              | defined_value
                              | number_form
                              | name_form)
    obj_id_components_list = OneOrMore(obj_id_components)
    object_identifier_value = ((Suppress(left_brace)
                                + obj_id_components_list
                                + Suppress(right_brace))
                               | (Suppress(left_brace)
                                  + defined_value
                                  + obj_id_components_list
                                  + Suppress(right_brace)))

    object_identifier_type = (OBJECT_IDENTIFIER
                              + Optional(left_parenthesis
                                         + delimitedList(word, delim='|')
                                         + right_parenthesis))
    object_identifier_type.setName('OBJECT IDENTIFIER')

    # X.680: 30. Notation for tagged types
    tagged_value = NoMatch()

    # X.680: 29. Notation for selection types

    # X.680: 28. Notation for the choice types
    alternative_type_list = delimitedList(named_type)
    extension_addition_alternatives_group = (Suppress(left_version_brackets)
                                             + Suppress(version_number)
                                             - alternative_type_list
                                             - Suppress(right_version_brackets))
    extension_addition_alternative = (extension_addition_alternatives_group
                                      | named_type)
    extension_addition_alternatives_list = delimitedList(extension_addition_alternative)
    extension_addition_alternatives = Optional(Suppress(comma)
                                               + extension_addition_alternatives_list)
    root_alternative_type_list = alternative_type_list
    alternative_type_lists = (root_alternative_type_list
                              + Optional(Suppress(comma)
                                         + extension_and_exception
                                         + extension_addition_alternatives
                                         + optional_extension_marker))
    choice_type = (CHOICE
                   - left_brace
                   + Group(alternative_type_lists)
                   - right_brace)
    choice_type.setName('CHOICE')
    choice_value = (identifier + colon + value)

    # X.680: 27. Notation for the set-of types
    set_of_value = NoMatch()
    set_of_type = (SET
                   + Group(Optional(size))
                   + OF
                   + Optional(Suppress(identifier))
                   - tag
                   - type_)
    set_of_type.setName('SET OF')

    # X.680: 26. Notation for the set types
    set_value = NoMatch()
    set_type = (SET
                + left_brace
                + Group(Optional(component_type_lists
                                 | (extension_and_exception
                                    + optional_extension_marker)))
                - right_brace)
    set_type.setName('SET')

    # X.680: 25. Notation for the sequence-of types
    sequence_of_value = NoMatch()
    sequence_of_type = (SEQUENCE
                        + Group(Optional(size_paren))
                        + OF
                        + Optional(Suppress(identifier))
                        - tag
                        - type_)
    sequence_of_type.setName('SEQUENCE OF')

    # X.680: 24. Notation for the sequence types
    component_value_list = delimitedList(named_value)
    sequence_value <<= (left_brace
                        + Optional(component_value_list)
                        + right_brace)
    component_type = Group(named_type
                           + Group(Optional(OPTIONAL
                                            | (DEFAULT + value)))
                           | (COMPONENTS_OF - type_))
    version_number <<= Optional(number + Suppress(colon))
    extension_addition_group = (Suppress(left_version_brackets)
                                + Suppress(version_number)
                                + delimitedList(component_type)
                                + Suppress(right_version_brackets))
    extension_and_exception <<= (ellipsis + Optional(exception_spec))
    extension_addition = (component_type | extension_addition_group)
    extension_addition_list = delimitedList(extension_addition)
    extension_additions = Optional(Suppress(comma) + extension_addition_list)
    extension_end_marker = (Suppress(comma) + ellipsis)
    optional_extension_marker <<= Optional(Suppress(comma) + ellipsis)
    component_type_list = delimitedList(component_type)
    root_component_type_list = component_type_list
    component_type_lists <<= ((root_component_type_list
                               + Optional(Suppress(comma)
                                          + extension_and_exception
                                          + extension_additions
                                          + ((extension_end_marker
                                              + Suppress(comma)
                                              + root_component_type_list)
                                             | optional_extension_marker)))
                              | (extension_and_exception
                                 + extension_additions
                                 + ((extension_end_marker
                                     + Suppress(comma)
                                     + root_component_type_list)
                                    | optional_extension_marker)))
    sequence_type = (SEQUENCE
                     - left_brace
                     + Group(Optional(component_type_lists
                                      | (extension_and_exception
                                         + optional_extension_marker)))
                     - right_brace)
    sequence_type.setName('SEQUENCE')

    # X.680: 23. Notation for the null type
    null_value = NULL
    null_type = NULL

    # X.680: 22. Notation for the octetstring type
    octet_string_value = (bstring
                          | hstring
                          | (CONTAINING + value))
    octet_string_type = OCTET_STRING
    octet_string_type.setName('OCTET STRING')

    # X.680: 21. Notation for the bitstring type
    bit_string_type = (BIT_STRING
                       + Group(Optional(left_brace
                                        + Group(delimitedList(word
                                                              + left_parenthesis
                                                              + word
                                                              + right_parenthesis))
                                        + right_brace)))
    bit_string_type.setName('BIT STRING')
    bit_string_value = Tag('BitStringValue',
                           bstring
                           | hstring
                           | Tag('IdentifierList',
                                 Suppress(left_brace)
                                 + Optional(identifier_list)
                                 + Suppress(right_brace))
                           | (CONTAINING - value))

    # X.680: 20. Notation for the real type
    special_real_value = (PLUS_INFINITY
                          | MINUS_INFINITY)
    numeric_real_value = (real_number
                          | sequence_value)
    real_value = (numeric_real_value
                  | special_real_value)
    real_type = REAL
    real_type.setName('REAL')

    # X.680: 19. Notation for the enumerated type
    enumerated_value = identifier
    enumerated_type = (ENUMERATED
                       - left_brace
                       + Group(delimitedList(Group((word
                                                    + Optional(Suppress(left_parenthesis)
                                                               + word
                                                               + Suppress(right_parenthesis)))
                                                   | ellipsis)))
                       - right_brace)
    enumerated_type.setName('ENUMERATED')

    # X.680: 18. Notation for the integer type
    integer_value = (signed_number | identifier)
    signed_number <<= Combine(Optional('-') + number)
    named_number = (identifier
                    + left_parenthesis
                    + (signed_number | defined_value)
                    + right_parenthesis)
    named_number_list = delimitedList(named_number)
    integer_type = (INTEGER
                    + Group(Optional(left_brace
                                     + named_number_list
                                     + right_brace)))
    integer_type.setName('INTEGER')

    # X.680: 17. Notation for boolean type
    boolean_type = BOOLEAN
    boolean_value = (TRUE | FALSE)

    # X.680: 16. Definition of types and values
    named_value <<= (identifier + value)
    referenced_value <<= NoMatch().setName('"referencedValue" not implemented')
    builtin_value <<= (bit_string_value
                       | boolean_value
                       | character_string_value
                       | choice_value
                       | relative_oid_value
                       | sequence_value
                       # | embedded_pdv_value
                       | enumerated_value
                       # | external_value
                       # | instance_of_value
                       | real_value
                       | integer_value
                       | null_value
                       | object_identifier_value
                       # | octet_string_value
                       | sequence_of_value
                       # | set_value
                       # | set_of_value
                       | tagged_value)
    value <<= Group(object_class_field_value)
                    # | referenced_value
                    # | builtin_value)
    named_type <<= Group(identifier
                         - tag
                         - type_)
    referenced_type = defined_type
    referenced_type.setName('ReferencedType')
    builtin_type = (choice_type
                    | integer_type
                    | null_type
                    | real_type
                    | bit_string_type
                    | octet_string_type
                    | enumerated_type
                    | sequence_of_type
                    | sequence_type
                    | object_class_field_type
                    | set_of_type
                    | set_type
                    | object_identifier_type
                    | boolean_type
                    | character_string_type)
    type_ <<= Group((builtin_type
                     | any_defined_by_type
                     | referenced_type).setName('Type')
                    + Group(ZeroOrMore(constraint)))

    # X.680: 15. Assigning types and values
    type_reference <<= (NotAny(reserved_words)
                        + Regex(r'[A-Z][a-zA-Z0-9-]*'))
    value_reference <<= Regex(r'[a-z][a-zA-Z0-9-]*')
    value_set <<= NoMatch().setName('"valueSet" not implemented')
    parameterized_type_assignment = (type_reference
                                     + parameter_list
                                     - assign
                                     - tag
                                     - type_)
    parameterized_value_assignment = (value_reference
                                      + parameter_list
                                      - Group(type_)
                                      - Suppress(assign)
                                      - value)

    # X.680: 14. Notation to support references to ASN.1 components

    # X.680: 13. Referencing type and value definitions
    external_value_reference <<= (module_reference
                                  + dot
                                  + value_reference)
    external_type_reference <<= (module_reference
                                 + dot
                                 + type_reference)
    defined_type <<= (external_type_reference
                      | parameterized_type
                      | parameterized_value_set_type
                      | type_reference)
    defined_value <<= (external_value_reference
                       | parameterized_value
                       | value_reference)

    # X.680: 12. Module definition
    module_reference <<= (NotAny(reserved_words)
                          + Regex(r'[A-Z][a-zA-Z0-9-]*').setName('modulereference'))
    assigned_identifier = Suppress(Optional(object_identifier_value
                                            | (defined_value + ~(comma | FROM))))
    global_module_reference = (module_reference + assigned_identifier)
    reference <<= (type_reference
                   | value_reference
                   | object_class_reference
                   | object_reference
                   | object_set_reference)
    symbol = (parameterized_reference
              | reference)
    symbol_list = Group(delimitedList(symbol))
    symbols_from_module = (symbol_list
                           + FROM
                           + global_module_reference)
    symbols_imported = OneOrMore(Group(symbols_from_module))
    imports = Optional(Suppress(IMPORTS)
                       - symbols_imported
                       - Suppress(semi_colon))
    symbols_exported = OneOrMore(symbol_list)
    exports = Suppress(Optional(EXPORTS
                                - (ALL
                                   | (symbols_exported + semi_colon))))
    assignment = (parameterized_object_set_assignment
                  | parameterized_object_assignment
                  | parameterized_object_class_assignment
                  | parameterized_type_assignment
                  | parameterized_value_assignment)
    assignment_list = ZeroOrMore(assignment)
    module_body = (exports + imports + assignment_list)
    definitive_name_and_number_form = (identifier
                                       + Suppress(left_parenthesis)
                                       - definitive_number_form
                                       - Suppress(right_parenthesis))
    definitive_number_form <<= number
    definitive_obj_id_component = Group(definitive_name_and_number_form
                                        | name_form
                                        | definitive_number_form)
    definitive_obj_id_components_list = OneOrMore(definitive_obj_id_component)
    definitive_identifier = Group(Optional(Suppress(left_brace)
                                           - definitive_obj_id_components_list
                                           - Suppress(right_brace)))
    module_identifier = (module_reference
                         + definitive_identifier)
    tag_default = Group(Optional((AUTOMATIC | EXPLICIT | IMPLICIT) + TAGS))
    extension_default = Group(Optional(EXTENSIBILITY_IMPLIED))
    module_definition = (Group(module_identifier
                               - Suppress(DEFINITIONS)
                               + tag_default
                               + extension_default
                               - Suppress(assign)
                               - Suppress(BEGIN))
                         + Group(module_body)
                         - Suppress(END))

    # The whole specification.
    specification = OneOrMore(module_definition) + StringEnd()

    # Parse actions converting tokens to asn1tools representation.
    integer.setParseAction(convert_integer)
    signed_number.setParseAction(convert_integer)
    real_number.setParseAction(convert_real_number)
    bstring.setParseAction(convert_bstring)
    hstring.setParseAction(convert_hstring)
    value_range.setParseAction(convert_value_range)
    size_constraint.setParseAction(convert_size_constraint)
    permitted_alphabet.setParseAction(convert_permitted_alphabet)
    constraint.setParseAction(convert_constraint)
    module_body.setParseAction(convert_module_body)
    specification.setParseAction(convert_specification)
    module_definition.setParseAction(convert_module_definition)
    assignment_list.setParseAction(convert_assignment_list)
    imports.setParseAction(convert_imports)
    parameterized_object_set_assignment.setParseAction(
        convert_parameterized_object_set_assignment)
    parameterized_object_assignment.setParseAction(
        convert_parameterized_object_assignment)
    parameterized_object_class_assignment.setParseAction(
        convert_parameterized_object_class_assignment)
    parameterized_type_assignment.setParseAction(
        convert_parameterized_type_assignment)
    parameterized_value_assignment.setParseAction(
        convert_parameterized_value_assignment)
    sequence_type.setParseAction(convert_sequence_type)
    sequence_of_type.setParseAction(convert_sequence_of_type)
    set_type.setParseAction(convert_set_type)
    set_of_type.setParseAction(convert_set_of_type)
    integer_type.setParseAction(convert_integer_type)
    real_type.setParseAction(convert_real_type)
    boolean_type.setParseAction(convert_boolean_type)
    bit_string_type.setParseAction(convert_bit_string_type)
    octet_string_type.setParseAction(convert_octet_string_type)
    null_type.setParseAction(convert_null_type)
    object_identifier_type.setParseAction(convert_object_identifier_type)
    enumerated_type.setParseAction(convert_enumerated_type)
    choice_type.setParseAction(convert_choice_type)
    defined_type.setParseAction(convert_defined_type)
    character_string_type.setParseAction(convert_keyword_type)
    object_class_field_type.setParseAction(convert_keyword_type)
    any_defined_by_type.setParseAction(convert_any_defined_by_type)

    return specification


def ignore_comments(string):
    """Ignore comments in given string by replacing them with spaces. This
    reduces the parsing time by roughly a factor of two.

    """

    re_replace = re.compile(r'[^\n]')

    return re.sub(r"--([\s\S]*?(--|\n))",
                  lambda mo: re_replace.sub(' ', mo.group(0)),
                  string)


def parse_string(string):
    """Parse given ASN.1 specification string and return a dictionary of
    its contents.

    >>> with open('foo.asn') as fin:
    ...     foo = asn1tools.parse_string(fin.read())

    """

    grammar = create_grammar()

    try:
        string = ignore_comments(string)
        tokens = grammar.parseString(string).asList()
    except (ParseException, ParseSyntaxException) as e:
        raise ParseError("Invalid ASN.1 syntax at line {}, column {}: '{}': {}.".format(
            e.lineno,
            e.column,
            e.markInputline(),
            e.msg))

    return tokens[0]


def parse_files(filenames):
    """Parse given ASN.1 specification file(s) and return a dictionary of
    its/their contents.

    >>> foo = asn1tools.parse_files('foo.asn')

    """

    if isinstance(filenames, str):
        filenames = [filenames]

    string = ''

    for filename in filenames:
        with open(filename, 'r') as fin:
            string += fin.read()
            string += '\n'

    return parse_string(string)



"""The top level of the ASN.1 tools package contains commonly used
functions and classes, and the command line interface.

"""

import sys
import argparse
import binascii
import logging

from .compiler import compile_dict
from .compiler import compile_string
from .compiler import compile_files
from .compiler import pre_process_dict
from .parser import parse_string
from .parser import parse_files
from .parser import ParseError
from .errors import EncodeError
from .errors import DecodeError
from .errors import CompileError


__author__ = 'Erik Moqvist'
__version__ = '0.33.0'


def print_list(list_, indent):
    for i, element in enumerate(list_):
        print('{}[{}]:'.format(indent * ' ', i))
        if isinstance(element, list):
            print_list(element, indent + 2)
        elif isinstance(element, dict):
            print_dict(element, indent + 2)
        elif isinstance(element, tuple):
            decoded = binascii.hexlify(element[0]).decode('ascii')
            print('{}"{}"'.format(indent * ' ', decoded))
        elif isinstance(element, bytearray):
            decoded = binascii.hexlify(element).decode('ascii')
            print("{}'{}'".format(indent * ' ', decoded))
        else:
            print('{}{}'.format(indent * ' ', element))


def print_dict(dict_, indent=0):
    for key, value in dict_.items():
        if isinstance(value, list):
            print('{}{}:'.format(indent * ' ', key))
            print_list(value, indent + 2)
        elif isinstance(value, dict):
            print('{}{}:'.format(indent * ' ', key))
            print_dict(value, indent + 2)
        elif isinstance(value, tuple):
            decoded = binascii.hexlify(value[0]).decode('ascii')
            print('{}{}: "{}"'.format(indent * ' ', key, decoded))
        elif isinstance(value, bytearray):
            decoded = binascii.hexlify(value).decode('ascii')
            print("{}{}: '{}'".format(indent * ' ', key, decoded))
        else:
            print('{}{}: {}'.format(indent * ' ', key, value))


def _do_decode(args):
    specification = compile_files(args.specification, args.codec)

    try:
        encoded = binascii.unhexlify(args.hexstring)
    except Exception as e:
        raise TypeError("'{}': {}".format(args.hexstring, str(e)))

    print_dict(specification.decode(args.type, encoded))


def _main():
    parser = argparse.ArgumentParser(
        description='Various ASN.1 utilities.')

    parser.add_argument('-d', '--debug', action='store_true')
    parser.add_argument('-v', '--verbose',
                        choices=(0, 1, 2),
                        type=int,
                        default=1,
                        help=('Control the verbosity; disable(0), warning(1) '
                              'and debug(2). (default: 1).'))
    parser.add_argument('--version',
                        action='version',
                        version=__version__,
                        help='Print version information and exit.')

    # Workaround to make the subparser required in Python 3.
    subparsers = parser.add_subparsers(title='subcommands',
                                       dest='subcommand')
    subparsers.required = True

    # The 'decode' subparser.
    decode_parser = subparsers.add_parser(
        'decode',
        description='Decode given hextring and print it to standard output.')
    decode_parser.add_argument('-c', '--codec',
                               choices=('ber', 'der', 'jer', 'per', 'uper', 'xer'),
                               default='ber',
                               help='Codec (default: ber).')
    decode_parser.add_argument('specification',
                               nargs='+',
                               help='ASN.1 specification as one or more .asn files.')
    decode_parser.add_argument('type', help='Type to decode.')
    decode_parser.add_argument('hexstring', help='Hexstring to decode.')
    decode_parser.set_defaults(func=_do_decode)

    args = parser.parse_args()

    levels = [logging.CRITICAL, logging.WARNING, logging.DEBUG]
    level = levels[args.verbose]

    logging.basicConfig(format='%(levelname)s: %(message)s',
                        level=level)

    if args.debug:
        args.func(args)
    else:
        try:
            args.func(args)
        except BaseException as e:
            sys.exit('error: {}'.format(str(e)))



import sys
import unittest

try:
    from StringIO import StringIO
except ImportError:
    from io import StringIO

try:
    from unittest.mock import patch
except ImportError:
    from mock import patch

import asn1tools


class Asn1ToolsCommandLineTest(unittest.TestCase):

    maxDiff = None

    def test_command_line_help(self):
        argv = ['asn1tools', '--help']

        stdout = StringIO()

        with patch('sys.stdout', stdout):
            with patch('sys.argv', argv):
                with self.assertRaises(SystemExit):
                    asn1tools._main()

        expected_output = [
            "usage: asn1tools [-h] [-d] [-v {0,1,2}] [--version] {decode} ...",
            "",
            "Various ASN.1 utilities.",
            "",
            "optional arguments:",
            "  -h, --help            show this help message and exit",
            "  -d, --debug",
            "  -v {0,1,2}, --verbose {0,1,2}",
            "                        Control the verbosity; ",
            "disable(0),",
            "warning(1)",
            "and",
            "debug(2).",
            "(default: 1).",
            "  --version             Print version information and exit.",
            "",
            "subcommands:",
            "  {decode}"
        ]

        for line in expected_output:
            self.assertIn(line, stdout.getvalue())

    def test_command_line_decode_help(self):
        argv = ['asn1tools', 'decode', '--help']

        stdout = StringIO()

        with patch('sys.stdout', stdout):
            with patch('sys.argv', argv):
                with self.assertRaises(SystemExit):
                    asn1tools._main()

        expected_output = [
            "asn1tools decode [-h] [-c {ber,der,jer,per,uper,xer}]",
            "                        specification [specification ...] type hexstring",
            "",
            "Decode given hextring and print it to standard output.",
            "",
            "positional arguments:",
            "  specification         ASN.1 specification as one or more .asn files.",
            "  type                  Type to decode.",
            "  hexstring             Hexstring to decode.",
            "",
            "optional arguments:",
            "  -h, --help            show this help message and exit",
            "  -c {ber,der,jer,per,uper,xer}, --codec {ber,der,jer,per,uper,xer}",
            "                        Codec (default: ber)."
        ]

        for line in expected_output:
            self.assertIn(line, stdout.getvalue())

    def test_command_line_decode_ber_foo_question(self):
        argv = ['asn1tools',
                'decode',
                'tests/files/foo.asn',
                'Question',
                '300e0201011609497320312b313d333f']

        stdout = StringIO()

        with patch('sys.stdout', stdout):
            with patch('sys.argv', argv):
                asn1tools._main()

        self.assertIn('id: 1', stdout.getvalue())
        self.assertIn('question: Is 1+1=3?', stdout.getvalue())

    def test_command_line_decode_uper_foo_question(self):
        argv = ['asn1tools',
                'decode',
                '--codec', 'uper',
                'tests/files/foo.asn',
                'Question',
                '01010993cd03156c5eb37e']

        stdout = StringIO()

        with patch('sys.stdout', stdout):
            with patch('sys.argv', argv):
                asn1tools._main()

        self.assertIn('id: 1', stdout.getvalue())
        self.assertIn('question: Is 1+1=3?', stdout.getvalue())

    def test_command_line_decode_ber_rrc_bcch_dl_sch_message(self):
        argv = ['asn1tools',
                'decode',
                'tests/files/ietf/rfc1155.asn',
                'tests/files/ietf/rfc1157.asn',
                'Message',
                '30819f02010004067075626c6963a3819102013c020100020100308185302206'
                '122b06010401817d08330a0201070a86deb735040c3137322e33312e31392e37'
                '33301706122b06010401817d08330a0201050a86deb960020102302306122b06'
                '010401817d08330a0201070a86deb736040d3235352e3235352e3235352e3030'
                '2106122b06010401817d08330a0201070a86deb738040b3137322e33312e3139'
                '2e32']

        stdout = StringIO()

        with patch('sys.stdout', stdout):
            with patch('sys.argv', argv):
                asn1tools._main()

        expected_output = [
            "community: '7075626c6963'",
            "data:",
            "  set-request:",
            "    variable-bindings:",
            "      [0]:",
            "        value:",
            "          simple:",
            "            string: '3137322e33312e31392e3733'",
            "        name: 1.3.6.1.4.1.253.8.51.10.2.1.7.10.14130101",
            "      [1]:",
            "        value:",
            "          simple:",
            "            number: 2",
            "        name: 1.3.6.1.4.1.253.8.51.10.2.1.5.10.14130400",
            "      [2]:",
            "        value:",
            "          simple:",
            "            string: '3235352e3235352e3235352e30'",
            "        name: 1.3.6.1.4.1.253.8.51.10.2.1.7.10.14130102",
            "      [3]:",
            "        value:",
            "          simple:",
            "            string: '3137322e33312e31392e32'",
            "        name: 1.3.6.1.4.1.253.8.51.10.2.1.7.10.14130104",
            "    error-index: 0",
            "    error-status: 0",
            "    request-id: 60",
            "version: 0"
        ]

        for line in expected_output:
            self.assertIn(line, stdout.getvalue())

    def test_command_line_decode_bad_type_name(self):
        argv = ['asn1tools',
                'decode',
                'tests/files/foo.asn',
                'Question2',
                '01010993cd03156c5eb37e']

        stdout = StringIO()

        with patch('sys.stdout', stdout):
            with patch('sys.argv', argv):
                with self.assertRaises(SystemExit) as cm:
                    asn1tools._main()

                self.assertEqual(
                    str(cm.exception),
                    "error: type 'Question2' not found in types dictionary")

    def test_command_line_decode_bad_data(self):
        argv = ['asn1tools',
                'decode',
                'tests/files/foo.asn',
                'Question',
                '012']

        stdout = StringIO()

        with patch('sys.stdout', stdout):
            with patch('sys.argv', argv):
                with self.assertRaises(SystemExit) as cm:
                    asn1tools._main()

                self.assertEqual(str(cm.exception),
                                 "error: '012': Odd-length string")


if __name__ == '__main__':
    unittest.main()



EXPECTED = {'PKIXAttributeCertificate': {'extensibility-implied': False,
                              'imports': {'PKIX1Explicit88': ['Attribute',
                                                              'AlgorithmIdentifier',
                                                              'CertificateSerialNumber',
                                                              'Extensions',
                                                              'UniqueIdentifier',
                                                              'id-pkix',
                                                              'id-pe',
                                                              'id-kp',
                                                              'id-ad',
                                                              'id-at'],
                                          'PKIX1Implicit88': ['GeneralName',
                                                              'GeneralNames',
                                                              'id-ce']},
                              'object-classes': {},
                              'object-sets': {},
                              'tags': 'IMPLICIT',
                              'types': {'AAControls': {'members': [{'name': 'pathLenConstraint',
                                                                    'optional': True,
                                                                    'restricted-to': [(0,
                                                                                       'MAX')],
                                                                    'type': 'INTEGER'},
                                                                   {'name': 'permittedAttrs',
                                                                    'optional': True,
                                                                    'tag': {'number': 0},
                                                                    'type': 'AttrSpec'},
                                                                   {'name': 'excludedAttrs',
                                                                    'optional': True,
                                                                    'tag': {'number': 1},
                                                                    'type': 'AttrSpec'},
                                                                   {'default': 'TRUE',
                                                                    'name': 'permitUnSpecified',
                                                                    'type': 'BOOLEAN'}],
                                                       'type': 'SEQUENCE'},
                                        'ACClearAttrs': {'members': [{'name': 'acIssuer',
                                                                      'type': 'GeneralName'},
                                                                     {'name': 'acSerial',
                                                                      'type': 'INTEGER'},
                                                                     {'element': {'type': 'Attribute'},
                                                                      'name': 'attrs',
                                                                      'type': 'SEQUENCE '
                                                                              'OF'}],
                                                         'type': 'SEQUENCE'},
                                        'AttCertIssuer': {'members': [{'name': 'v1Form',
                                                                       'type': 'GeneralNames'},
                                                                      {'name': 'v2Form',
                                                                       'tag': {'number': 0},
                                                                       'type': 'V2Form'}],
                                                          'type': 'CHOICE'},
                                        'AttCertValidityPeriod': {'members': [{'name': 'notBeforeTime',
                                                                               'type': 'GeneralizedTime'},
                                                                              {'name': 'notAfterTime',
                                                                               'type': 'GeneralizedTime'}],
                                                                  'type': 'SEQUENCE'},
                                        'AttCertVersion': {'type': 'INTEGER'},
                                        'AttrSpec': {'element': {'type': 'OBJECT '
                                                                         'IDENTIFIER'},
                                                     'type': 'SEQUENCE OF'},
                                        'AttributeCertificate': {'members': [{'name': 'acinfo',
                                                                              'type': 'AttributeCertificateInfo'},
                                                                             {'name': 'signatureAlgorithm',
                                                                              'type': 'AlgorithmIdentifier'},
                                                                             {'name': 'signatureValue',
                                                                              'type': 'BIT '
                                                                                      'STRING'}],
                                                                 'type': 'SEQUENCE'},
                                        'AttributeCertificateInfo': {'members': [{'name': 'version',
                                                                                  'type': 'AttCertVersion'},
                                                                                 {'name': 'holder',
                                                                                  'type': 'Holder'},
                                                                                 {'name': 'issuer',
                                                                                  'type': 'AttCertIssuer'},
                                                                                 {'name': 'signature',
                                                                                  'type': 'AlgorithmIdentifier'},
                                                                                 {'name': 'serialNumber',
                                                                                  'type': 'CertificateSerialNumber'},
                                                                                 {'name': 'attrCertValidityPeriod',
                                                                                  'type': 'AttCertValidityPeriod'},
                                                                                 {'element': {'type': 'Attribute'},
                                                                                  'name': 'attributes',
                                                                                  'type': 'SEQUENCE '
                                                                                          'OF'},
                                                                                 {'name': 'issuerUniqueID',
                                                                                  'optional': True,
                                                                                  'type': 'UniqueIdentifier'},
                                                                                 {'name': 'extensions',
                                                                                  'optional': True,
                                                                                  'type': 'Extensions'}],
                                                                     'type': 'SEQUENCE'},
                                        'ClassList': {'type': 'BIT STRING'},
                                        'Clearance': {'members': [{'name': 'policyId',
                                                                   'tag': {'number': 0},
                                                                   'type': 'OBJECT '
                                                                           'IDENTIFIER'},
                                                                  {'default': ['unclassified'],
                                                                   'name': 'classList',
                                                                   'tag': {'number': 1},
                                                                   'type': 'ClassList'},
                                                                  {'element': {'type': 'SecurityCategory'},
                                                                   'name': 'securityCategories',
                                                                   'optional': True,
                                                                   'tag': {'number': 2},
                                                                   'type': 'SET '
                                                                           'OF'}],
                                                      'type': 'SEQUENCE'},
                                        'Holder': {'members': [{'name': 'baseCertificateID',
                                                                'optional': True,
                                                                'tag': {'number': 0},
                                                                'type': 'IssuerSerial'},
                                                               {'name': 'entityName',
                                                                'optional': True,
                                                                'tag': {'number': 1},
                                                                'type': 'GeneralNames'},
                                                               {'name': 'objectDigestInfo',
                                                                'optional': True,
                                                                'tag': {'number': 2},
                                                                'type': 'ObjectDigestInfo'}],
                                                   'type': 'SEQUENCE'},
                                        'IetfAttrSyntax': {'members': [{'name': 'policyAuthority',
                                                                        'optional': True,
                                                                        'tag': {'number': 0},
                                                                        'type': 'GeneralNames'},
                                                                       {'element': {'members': [{'name': 'octets',
                                                                                                 'type': 'OCTET '
                                                                                                         'STRING'},
                                                                                                {'name': 'oid',
                                                                                                 'type': 'OBJECT '
                                                                                                         'IDENTIFIER'},
                                                                                                {'name': 'string',
                                                                                                 'type': 'UTF8String'}],
                                                                                    'type': 'CHOICE'},
                                                                        'name': 'values',
                                                                        'type': 'SEQUENCE '
                                                                                'OF'}],
                                                           'type': 'SEQUENCE'},
                                        'IssuerSerial': {'members': [{'name': 'issuer',
                                                                      'type': 'GeneralNames'},
                                                                     {'name': 'serial',
                                                                      'type': 'CertificateSerialNumber'},
                                                                     {'name': 'issuerUID',
                                                                      'optional': True,
                                                                      'type': 'UniqueIdentifier'}],
                                                         'type': 'SEQUENCE'},
                                        'ObjectDigestInfo': {'members': [{'name': 'digestedObjectType',
                                                                          'type': 'ENUMERATED',
                                                                          'values': {0: 'publicKey',
                                                                                     1: 'publicKeyCert',
                                                                                     2: 'otherObjectTypes'}},
                                                                         {'name': 'otherObjectTypeID',
                                                                          'optional': True,
                                                                          'type': 'OBJECT '
                                                                                  'IDENTIFIER'},
                                                                         {'name': 'digestAlgorithm',
                                                                          'type': 'AlgorithmIdentifier'},
                                                                         {'name': 'objectDigest',
                                                                          'type': 'BIT '
                                                                                  'STRING'}],
                                                             'type': 'SEQUENCE'},
                                        'ProxyInfo': {'element': {'type': 'Targets'},
                                                      'type': 'SEQUENCE OF'},
                                        'RoleSyntax': {'members': [{'name': 'roleAuthority',
                                                                    'optional': True,
                                                                    'tag': {'number': 0},
                                                                    'type': 'GeneralNames'},
                                                                   {'name': 'roleName',
                                                                    'tag': {'number': 1},
                                                                    'type': 'GeneralName'}],
                                                       'type': 'SEQUENCE'},
                                        'SecurityCategory': {'members': [{'name': 'type',
                                                                          'tag': {'kind': 'IMPLICIT',
                                                                                  'number': 0},
                                                                          'type': 'OBJECT '
                                                                                  'IDENTIFIER'},
                                                                         {'choices': {},
                                                                          'name': 'value',
                                                                          'tag': {'number': 1},
                                                                          'type': 'ANY '
                                                                                  'DEFINED '
                                                                                  'BY',
                                                                          'value': 'type'}],
                                                             'type': 'SEQUENCE'},
                                        'SvceAuthInfo': {'members': [{'name': 'service',
                                                                      'type': 'GeneralName'},
                                                                     {'name': 'ident',
                                                                      'type': 'GeneralName'},
                                                                     {'name': 'authInfo',
                                                                      'optional': True,
                                                                      'type': 'OCTET '
                                                                              'STRING'}],
                                                         'type': 'SEQUENCE'},
                                        'Target': {'members': [{'name': 'targetName',
                                                                'tag': {'number': 0},
                                                                'type': 'GeneralName'},
                                                               {'name': 'targetGroup',
                                                                'tag': {'number': 1},
                                                                'type': 'GeneralName'},
                                                               {'name': 'targetCert',
                                                                'tag': {'number': 2},
                                                                'type': 'TargetCert'}],
                                                   'type': 'CHOICE'},
                                        'TargetCert': {'members': [{'name': 'targetCertificate',
                                                                    'type': 'IssuerSerial'},
                                                                   {'name': 'targetName',
                                                                    'optional': True,
                                                                    'type': 'GeneralName'},
                                                                   {'name': 'certDigestInfo',
                                                                    'optional': True,
                                                                    'type': 'ObjectDigestInfo'}],
                                                       'type': 'SEQUENCE'},
                                        'Targets': {'element': {'type': 'Target'},
                                                    'type': 'SEQUENCE OF'},
                                        'V2Form': {'members': [{'name': 'issuerName',
                                                                'optional': True,
                                                                'type': 'GeneralNames'},
                                                               {'name': 'baseCertificateID',
                                                                'optional': True,
                                                                'tag': {'number': 0},
                                                                'type': 'IssuerSerial'},
                                                               {'name': 'objectDigestInfo',
                                                                'optional': True,
                                                                'tag': {'number': 1},
                                                                'type': 'ObjectDigestInfo'}],
                                                   'type': 'SEQUENCE'}},
                              'values': {'id-aca': {'type': 'OBJECT IDENTIFIER',
                                                    'value': ['id-pkix', 10]},
                                         'id-aca-accessIdentity': {'type': 'OBJECT '
                                                                           'IDENTIFIER',
                                                                   'value': ['id-aca',
                                                                             2]},
                                         'id-aca-authenticationInfo': {'type': 'OBJECT '
                                                                               'IDENTIFIER',
                                                                       'value': ['id-aca',
                                                                                 1]},
                                         'id-aca-chargingIdentity': {'type': 'OBJECT '
                                                                             'IDENTIFIER',
                                                                     'value': ['id-aca',
                                                                               3]},
                                         'id-aca-encAttrs': {'type': 'OBJECT '
                                                                     'IDENTIFIER',
                                                             'value': ['id-aca',
                                                                       6]},
                                         'id-aca-group': {'type': 'OBJECT '
                                                                  'IDENTIFIER',
                                                          'value': ['id-aca',
                                                                    4]},
                                         'id-at-clearance': {'type': 'OBJECT '
                                                                     'IDENTIFIER',
                                                             'value': [('joint-iso-ccitt',
                                                                        2),
                                                                       ('ds',
                                                                        5),
                                                                       ('module',
                                                                        1),
                                                                       ('selected-attribute-types',
                                                                        5),
                                                                       ('clearance',
                                                                        55)]},
                                         'id-at-role': {'type': 'OBJECT '
                                                                'IDENTIFIER',
                                                        'value': ['id-at', 72]},
                                         'id-ce-targetInformation': {'type': 'OBJECT '
                                                                             'IDENTIFIER',
                                                                     'value': ['id-ce',
                                                                               55]},
                                         'id-pe-aaControls': {'type': 'OBJECT '
                                                                      'IDENTIFIER',
                                                              'value': ['id-pe',
                                                                        6]},
                                         'id-pe-ac-auditIdentity': {'type': 'OBJECT '
                                                                            'IDENTIFIER',
                                                                    'value': ['id-pe',
                                                                              4]},
                                         'id-pe-ac-proxying': {'type': 'OBJECT '
                                                                       'IDENTIFIER',
                                                               'value': ['id-pe',
                                                                         10]}}}}


EXPECTED = {'PKIX1Algorithms88': {'extensibility-implied': False,
                       'imports': {},
                       'object-classes': {},
                       'object-sets': {},
                       'tags': 'EXPLICIT',
                       'types': {'Characteristic-two': {'members': [{'name': 'm',
                                                                     'type': 'INTEGER'},
                                                                    {'name': 'basis',
                                                                     'type': 'OBJECT '
                                                                             'IDENTIFIER'},
                                                                    {'choices': {},
                                                                     'name': 'parameters',
                                                                     'type': 'ANY '
                                                                             'DEFINED '
                                                                             'BY',
                                                                     'value': 'basis'}],
                                                        'type': 'SEQUENCE'},
                                 'Curve': {'members': [{'name': 'a',
                                                        'type': 'FieldElement'},
                                                       {'name': 'b',
                                                        'type': 'FieldElement'},
                                                       {'name': 'seed',
                                                        'optional': True,
                                                        'type': 'BIT STRING'}],
                                           'type': 'SEQUENCE'},
                                 'DHPublicKey': {'type': 'INTEGER'},
                                 'DSAPublicKey': {'type': 'INTEGER'},
                                 'DomainParameters': {'members': [{'name': 'p',
                                                                   'type': 'INTEGER'},
                                                                  {'name': 'g',
                                                                   'type': 'INTEGER'},
                                                                  {'name': 'q',
                                                                   'type': 'INTEGER'},
                                                                  {'name': 'j',
                                                                   'optional': True,
                                                                   'type': 'INTEGER'},
                                                                  {'name': 'validationParms',
                                                                   'optional': True,
                                                                   'type': 'ValidationParms'}],
                                                      'type': 'SEQUENCE'},
                                 'Dss-Parms': {'members': [{'name': 'p',
                                                            'type': 'INTEGER'},
                                                           {'name': 'q',
                                                            'type': 'INTEGER'},
                                                           {'name': 'g',
                                                            'type': 'INTEGER'}],
                                               'type': 'SEQUENCE'},
                                 'Dss-Sig-Value': {'members': [{'name': 'r',
                                                                'type': 'INTEGER'},
                                                               {'name': 's',
                                                                'type': 'INTEGER'}],
                                                   'type': 'SEQUENCE'},
                                 'ECDSA-Sig-Value': {'members': [{'name': 'r',
                                                                  'type': 'INTEGER'},
                                                                 {'name': 's',
                                                                  'type': 'INTEGER'}],
                                                     'type': 'SEQUENCE'},
                                 'ECPVer': {'type': 'INTEGER'},
                                 'ECParameters': {'members': [{'name': 'version',
                                                               'type': 'ECPVer'},
                                                              {'name': 'fieldID',
                                                               'type': 'FieldID'},
                                                              {'name': 'curve',
                                                               'type': 'Curve'},
                                                              {'name': 'base',
                                                               'type': 'ECPoint'},
                                                              {'name': 'order',
                                                               'type': 'INTEGER'},
                                                              {'name': 'cofactor',
                                                               'optional': True,
                                                               'type': 'INTEGER'}],
                                                  'type': 'SEQUENCE'},
                                 'ECPoint': {'type': 'OCTET STRING'},
                                 'EcpkParameters': {'members': [{'name': 'ecParameters',
                                                                 'type': 'ECParameters'},
                                                                {'name': 'namedCurve',
                                                                 'type': 'OBJECT '
                                                                         'IDENTIFIER'},
                                                                {'name': 'implicitlyCA',
                                                                 'type': 'NULL'}],
                                                    'type': 'CHOICE'},
                                 'FieldElement': {'type': 'OCTET STRING'},
                                 'FieldID': {'members': [{'name': 'fieldType',
                                                          'type': 'OBJECT '
                                                                  'IDENTIFIER'},
                                                         {'choices': {},
                                                          'name': 'parameters',
                                                          'type': 'ANY DEFINED '
                                                                  'BY',
                                                          'value': 'fieldType'}],
                                             'type': 'SEQUENCE'},
                                 'KEA-Parms-Id': {'type': 'OCTET STRING'},
                                 'Pentanomial': {'members': [{'name': 'k1',
                                                              'type': 'INTEGER'},
                                                             {'name': 'k2',
                                                              'type': 'INTEGER'},
                                                             {'name': 'k3',
                                                              'type': 'INTEGER'}],
                                                 'type': 'SEQUENCE'},
                                 'Prime-p': {'type': 'INTEGER'},
                                 'RSAPublicKey': {'members': [{'name': 'modulus',
                                                               'type': 'INTEGER'},
                                                              {'name': 'publicExponent',
                                                               'type': 'INTEGER'}],
                                                  'type': 'SEQUENCE'},
                                 'Trinomial': {'type': 'INTEGER'},
                                 'ValidationParms': {'members': [{'name': 'seed',
                                                                  'type': 'BIT '
                                                                          'STRING'},
                                                                 {'name': 'pgenCounter',
                                                                  'type': 'INTEGER'}],
                                                     'type': 'SEQUENCE'}},
                       'values': {'ansi-X9-62': {'type': 'OBJECT IDENTIFIER',
                                                 'value': [('iso', 1),
                                                           ('member-body', 2),
                                                           ('us', 840),
                                                           10045]},
                                  'c-TwoCurve': {'type': 'OBJECT IDENTIFIER',
                                                 'value': ['ellipticCurve',
                                                           ('characteristicTwo',
                                                            0)]},
                                  'c2onb191v4': {'type': 'OBJECT IDENTIFIER',
                                                 'value': ['c-TwoCurve', 8]},
                                  'c2onb191v5': {'type': 'OBJECT IDENTIFIER',
                                                 'value': ['c-TwoCurve', 9]},
                                  'c2onb239v4': {'type': 'OBJECT IDENTIFIER',
                                                 'value': ['c-TwoCurve', 14]},
                                  'c2onb239v5': {'type': 'OBJECT IDENTIFIER',
                                                 'value': ['c-TwoCurve', 15]},
                                  'c2pnb163v1': {'type': 'OBJECT IDENTIFIER',
                                                 'value': ['c-TwoCurve', 1]},
                                  'c2pnb163v2': {'type': 'OBJECT IDENTIFIER',
                                                 'value': ['c-TwoCurve', 2]},
                                  'c2pnb163v3': {'type': 'OBJECT IDENTIFIER',
                                                 'value': ['c-TwoCurve', 3]},
                                  'c2pnb176w1': {'type': 'OBJECT IDENTIFIER',
                                                 'value': ['c-TwoCurve', 4]},
                                  'c2pnb208w1': {'type': 'OBJECT IDENTIFIER',
                                                 'value': ['c-TwoCurve', 10]},
                                  'c2pnb272w1': {'type': 'OBJECT IDENTIFIER',
                                                 'value': ['c-TwoCurve', 16]},
                                  'c2pnb304w1': {'type': 'OBJECT IDENTIFIER',
                                                 'value': ['c-TwoCurve', 17]},
                                  'c2pnb368w1': {'type': 'OBJECT IDENTIFIER',
                                                 'value': ['c-TwoCurve', 19]},
                                  'c2tnb191v1': {'type': 'OBJECT IDENTIFIER',
                                                 'value': ['c-TwoCurve', 5]},
                                  'c2tnb191v2': {'type': 'OBJECT IDENTIFIER',
                                                 'value': ['c-TwoCurve', 6]},
                                  'c2tnb191v3': {'type': 'OBJECT IDENTIFIER',
                                                 'value': ['c-TwoCurve', 7]},
                                  'c2tnb239v1': {'type': 'OBJECT IDENTIFIER',
                                                 'value': ['c-TwoCurve', 11]},
                                  'c2tnb239v2': {'type': 'OBJECT IDENTIFIER',
                                                 'value': ['c-TwoCurve', 12]},
                                  'c2tnb239v3': {'type': 'OBJECT IDENTIFIER',
                                                 'value': ['c-TwoCurve', 13]},
                                  'c2tnb359v1': {'type': 'OBJECT IDENTIFIER',
                                                 'value': ['c-TwoCurve', 18]},
                                  'c2tnb431r1': {'type': 'OBJECT IDENTIFIER',
                                                 'value': ['c-TwoCurve', 20]},
                                  'characteristic-two-field': {'type': 'OBJECT '
                                                                       'IDENTIFIER',
                                                               'value': ['id-fieldType',
                                                                         2]},
                                  'dhpublicnumber': {'type': 'OBJECT '
                                                             'IDENTIFIER',
                                                     'value': [('iso', 1),
                                                               ('member-body',
                                                                2),
                                                               ('us', 840),
                                                               ('ansi-x942',
                                                                10046),
                                                               ('number-type',
                                                                2),
                                                               1]},
                                  'ecdsa-with-SHA1': {'type': 'OBJECT '
                                                              'IDENTIFIER',
                                                      'value': ['id-ecSigType',
                                                                1]},
                                  'ellipticCurve': {'type': 'OBJECT IDENTIFIER',
                                                    'value': ['ansi-X9-62',
                                                              ('curves', 3)]},
                                  'gnBasis': {'type': 'OBJECT IDENTIFIER',
                                              'value': ['id-characteristic-two-basis',
                                                        1]},
                                  'id-characteristic-two-basis': {'type': 'OBJECT '
                                                                          'IDENTIFIER',
                                                                  'value': ['characteristic-two-field',
                                                                            ('basisType',
                                                                             3)]},
                                  'id-dsa': {'type': 'OBJECT IDENTIFIER',
                                             'value': [('iso', 1),
                                                       ('member-body', 2),
                                                       ('us', 840),
                                                       ('x9-57', 10040),
                                                       ('x9algorithm', 4),
                                                       1]},
                                  'id-dsa-with-sha1': {'type': 'OBJECT '
                                                               'IDENTIFIER',
                                                       'value': [('iso', 1),
                                                                 ('member-body',
                                                                  2),
                                                                 ('us', 840),
                                                                 ('x9-57',
                                                                  10040),
                                                                 ('x9algorithm',
                                                                  4),
                                                                 3]},
                                  'id-ecPublicKey': {'type': 'OBJECT '
                                                             'IDENTIFIER',
                                                     'value': ['id-publicKeyType',
                                                               1]},
                                  'id-ecSigType': {'type': 'OBJECT IDENTIFIER',
                                                   'value': ['ansi-X9-62',
                                                             ('signatures',
                                                              4)]},
                                  'id-fieldType': {'type': 'OBJECT IDENTIFIER',
                                                   'value': ['ansi-X9-62',
                                                             ('fieldType', 1)]},
                                  'id-keyExchangeAlgorithm': {'type': 'OBJECT '
                                                                      'IDENTIFIER',
                                                              'value': [2,
                                                                        16,
                                                                        840,
                                                                        1,
                                                                        101,
                                                                        2,
                                                                        1,
                                                                        1,
                                                                        22]},
                                  'id-publicKeyType': {'type': 'OBJECT '
                                                               'IDENTIFIER',
                                                       'value': ['ansi-X9-62',
                                                                 ('keyType',
                                                                  2)]},
                                  'id-sha1': {'type': 'OBJECT IDENTIFIER',
                                              'value': [('iso', 1),
                                                        ('identified-organization',
                                                         3),
                                                        ('oiw', 14),
                                                        ('secsig', 3),
                                                        ('algorithms', 2),
                                                        26]},
                                  'md2': {'type': 'OBJECT IDENTIFIER',
                                          'value': [('iso', 1),
                                                    ('member-body', 2),
                                                    ('us', 840),
                                                    ('rsadsi', 113549),
                                                    ('digestAlgorithm', 2),
                                                    2]},
                                  'md2WithRSAEncryption': {'type': 'OBJECT '
                                                                   'IDENTIFIER',
                                                           'value': ['pkcs-1',
                                                                     2]},
                                  'md5': {'type': 'OBJECT IDENTIFIER',
                                          'value': [('iso', 1),
                                                    ('member-body', 2),
                                                    ('us', 840),
                                                    ('rsadsi', 113549),
                                                    ('digestAlgorithm', 2),
                                                    5]},
                                  'md5WithRSAEncryption': {'type': 'OBJECT '
                                                                   'IDENTIFIER',
                                                           'value': ['pkcs-1',
                                                                     4]},
                                  'pkcs-1': {'type': 'OBJECT IDENTIFIER',
                                             'value': [('iso', 1),
                                                       ('member-body', 2),
                                                       ('us', 840),
                                                       ('rsadsi', 113549),
                                                       ('pkcs', 1),
                                                       1]},
                                  'ppBasis': {'type': 'OBJECT IDENTIFIER',
                                              'value': ['id-characteristic-two-basis',
                                                        3]},
                                  'prime-field': {'type': 'OBJECT IDENTIFIER',
                                                  'value': ['id-fieldType', 1]},
                                  'prime192v1': {'type': 'OBJECT IDENTIFIER',
                                                 'value': ['primeCurve', 1]},
                                  'prime192v2': {'type': 'OBJECT IDENTIFIER',
                                                 'value': ['primeCurve', 2]},
                                  'prime192v3': {'type': 'OBJECT IDENTIFIER',
                                                 'value': ['primeCurve', 3]},
                                  'prime239v1': {'type': 'OBJECT IDENTIFIER',
                                                 'value': ['primeCurve', 4]},
                                  'prime239v2': {'type': 'OBJECT IDENTIFIER',
                                                 'value': ['primeCurve', 5]},
                                  'prime239v3': {'type': 'OBJECT IDENTIFIER',
                                                 'value': ['primeCurve', 6]},
                                  'prime256v1': {'type': 'OBJECT IDENTIFIER',
                                                 'value': ['primeCurve', 7]},
                                  'primeCurve': {'type': 'OBJECT IDENTIFIER',
                                                 'value': ['ellipticCurve',
                                                           ('prime', 1)]},
                                  'rsaEncryption': {'type': 'OBJECT IDENTIFIER',
                                                    'value': ['pkcs-1', 1]},
                                  'sha1WithRSAEncryption': {'type': 'OBJECT '
                                                                    'IDENTIFIER',
                                                            'value': ['pkcs-1',
                                                                      5]},
                                  'tpBasis': {'type': 'OBJECT IDENTIFIER',
                                              'value': ['id-characteristic-two-basis',
                                                        2]}}}}